/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// Created by: Sergey Belovolov                                                         
// Created on: 23.06.04                                                                 
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

#ifndef _IP_SERVER_H
#define _IP_SERVER_H

#include "boost/shared_ptr.hpp"
#include "boost/thread/mutex.hpp"
#include "boost/thread/thread.hpp"

#include "ipstream/ipstream.h"

#include "boost_adopt/boost_adopt.h" 

#include <map>

namespace IPStream
{
	struct IPortProcessor;

	// Number of seconds for main thread to sleep (100 milliseconds)
	//static const unsigned int cuiMainSleep = 1;
	static const unsigned int cuiMainSleep = 100000000;

	// Lock type for Server
	typedef BMutex::scoped_lock SrvLock;

	// Data for thread (for servers realizations)
	struct THREAD_DATA
	{
		// Assotiated processor
		IPortProcessor* m_pPortProcessor;
		// Thread
		SPThread m_spThread;
		// Flag if we have to delete this thread
		volatile bool m_bHaveToDel;

		// Flag if we have to stop thread
		volatile bool m_bHaveToStop;

		// Domain at what socket will work
		EDomain m_EDomain;

		// Possible options for socket
		int m_iOptions;

		// Default constructor
		THREAD_DATA() : 
		m_pPortProcessor( 0 ), 
			m_bHaveToDel( false ),
			m_bHaveToStop( false ), 
			m_EDomain( D_INET ), 
			m_iOptions( 0 )
		{
			d_cerr( "THREAD_DATA creation." );
			d_cerr_sl( "Domain: " ); d_cerr( m_EDomain ); 
			d_cerr( "" );
		}

		~THREAD_DATA()
		{
			d_cerr( "THREAD_DATA destruction." );

			// If thread executes
			if ( !( m_bHaveToStop && m_bHaveToDel ) && m_spThread.get() )
			{
				d_cerr( "THREAD_DATA destructor tryes to join thread." );

				m_bHaveToDel = m_bHaveToStop = true;

				m_spThread -> join();
				m_spThread.reset();

				d_cerr( "THREAD_DATA destructor joined thread successfully." );
			}

			d_cerr( "" );
		}

	};

	typedef std::map < unsigned int, THREAD_DATA > ThreadMap;

	// Common interface template for TCP & UDP servers
	template < typename ServerClass >
	struct ITIPServer
	{
		private:
/**
* Section for organization interface for creation of the main thread
*/
		// Mutex for m_bStarted & m_bHaveToStop
		BMutex m_mStarted;

		// Flag which indicates if thread was started
		volatile bool m_bStarted;

		// Flag which indicates if thread have to stop
		volatile bool m_bHaveToStop;

		// Smartpointer to main thread
		SPThread m_spMainThread;

		// Section for safety (non-copyable class)
		ITIPServer( const ITIPServer& );
		ITIPServer& operator=( const ITIPServer& );
/**
 * End of section
 */
		protected:
		// Protocol of server
		// Don't forget to set it from child class constructor
		ETransportProtocol m_tpProtocol;
/**
 * Section for the data containers
 */
		// Mutex for m_mapThreads
		BMutex m_mContainer;

		// Map for storing accept threads for ports
		ThreadMap m_mapThreads;

#ifdef USE_SSL
		//SSL context object
		SPTR_CSSL_CTX m_ptrCSSL_CTX;
		bool isSSLused;
#endif

/**
 * End of section
 */
		public:
		// Main dispatcher thread 
		virtual void MainThread( void* )
		{
			d_cerr( "ITIPServer::MainThread creation." );

			// Main loop
			while ( !m_bHaveToStop )
			{
				// Temporary lock
				SrvLock TmpLock( m_mContainer );

				// Check if we have data in threads container
				if ( !m_mapThreads.empty() )
				{
					for ( ThreadMap::iterator tmciThread = m_mapThreads.begin(); 
						tmciThread != m_mapThreads.end(); )
					{
						// Check if we got thread that we've to stop or already finished
						if ( ( tmciThread -> second.m_bHaveToStop ) || 
							 ( tmciThread -> second.m_bHaveToDel ) )
						{
							// Check if thread is OK then join it
							if ( tmciThread -> second.m_spThread.get() )
							{
								// Wait until thread finish
								tmciThread -> second.m_spThread -> join();
								// Write NULL pointer there and delete pointed thread
								tmciThread -> second.m_spThread.reset();

								d_cerr( "ITIPServer::MainThread accept thread destruction." );
								d_cerr( "" );
							}

							if ( tmciThread -> second.m_bHaveToDel )
								// Remove this record
								m_mapThreads.erase( tmciThread++ );
							else
							{
								// For avoiding this branch 
								tmciThread -> second.m_bHaveToStop = false;

								++tmciThread;
							}

							// Go to next step
							continue;
						}
						else
						{
							// Check if thread is emty then create it
							if ( !( tmciThread -> second.m_spThread.get() ) )
							{
								ipstream ipsTmp;
#ifdef USE_SSL
								if ( isSSLused && m_ptrCSSL_CTX )
								{
									// ipstream creation
									ipsTmp = ipstream(
													 m_ptrCSSL_CTX,
													 tmciThread -> first, "",  
													 tmciThread -> second.m_iOptions,
													 tmciThread -> second.m_EDomain 
													 );
								}
								else
								{
									// ipstream creation
									ipsTmp = ipstream( tmciThread -> first, "", m_tpProtocol, 
													   tmciThread -> second.m_iOptions,
													   tmciThread -> second.m_EDomain 
													 );
								}
#else
								// ipstream creation
								ipsTmp = ipstream( tmciThread -> first, "", m_tpProtocol, 
												   tmciThread -> second.m_iOptions,
												   tmciThread -> second.m_EDomain 
												 );
#endif

								// If stream was successefully created
								if ( ipsTmp.good() )
								{
									// Create processing thread and assign it
									tmciThread -> 
										second.
										m_spThread.
										reset(
											 new boost::thread(
															  boost_adopt::class_member_2_thread_adapter 
															  < ServerClass, ipstream > 
															  ( ( ServerClass* ) this, 
																&ServerClass::AcceptThread, 
																ipsTmp ) ) );

									d_cerr( "ITIPServer::MainThread accept thread creation." );
									d_cerr( "" );
								}
								else
									d_cerr( "ITIPServer::MainThread ipstream isn't valid." );
							} // if ( !( tmciThread -> second.m_spThread.get() ) )

							++tmciThread;
						} // if - else
					} // for 
				} // if ( !m_mapThreads.empty() )

				// Release container lock
				TmpLock.unlock();

				boost_adopt::sleep( 0, cuiMainSleep );
			} // while ( !m_bHaveToStop )

			d_cerr( "ITIPServer::MainThread thread ended." );

			// Cleanup section
			// Temporary lock
			SrvLock TmpLock( m_mContainer );

			d_cerr( "ITIPServer::MainThread m_mContainer lock succeeded." );

			// All started accept threads destruction
			if ( !m_mapThreads.empty() )
			{
				for ( ThreadMap::iterator tmciThread = m_mapThreads.begin(); 
					tmciThread != m_mapThreads.end(); )
				{
					// Stop
					tmciThread -> second.m_bHaveToStop = true;

					// Check if thread is OK then join it
					if ( tmciThread -> second.m_spThread.get() )
					{
						d_cerr( "ITIPServer::MainThread joining thread." );

						TmpLock.unlock();

						tmciThread -> second.m_spThread -> join();
						// Write NULL pointer there and delete pointed thread

						TmpLock.lock();

						tmciThread -> second.m_spThread.reset();
					}

					if ( tmciThread -> second.m_bHaveToDel )
						// Remove this record
						m_mapThreads.erase( tmciThread++ );
					else
						++tmciThread;
				}
			} // if ( !m_mapThreads.empty() )

			d_cerr( "" );
		}


		// Processing thread (pure virtual!)
		virtual void AcceptThread( ipstream ) = 0;

		// Never use directly
		explicit ITIPServer() : 
		m_bStarted( false ), 
			m_bHaveToStop( false ), 
			m_tpProtocol( TP_RAW )
		{
#ifdef USE_SSL
			isSSLused = false;
#endif
			d_cerr( "ITIPServer creation." );
			d_cerr( "" );
		}

		virtual ~ITIPServer()
		{
			d_cerr( "ITIPServer destruction." );
			d_cerr( "" );

			// Stop main thread
			Stop();
		}

		// Start main dispatcher thread
		virtual bool Start()
		{
			d_cerr( "ITIPServer::Start call." );

			if ( !m_bStarted )
			{
				// Temporary lock
				SrvLock TmpLock( m_mStarted );

				// Double checking with lock
				if ( !m_bStarted )
				{
					// Set stop flag for main thread
					m_bHaveToStop = false;

					// Create main thread and assign it
					m_spMainThread.reset( new boost::thread(
														   boost_adopt::class_member_2_thread_adapter 
														   < ServerClass, void* > 
														   ( ( ServerClass* ) this, &ServerClass::MainThread, 
															 NULL ) ) ); 

					// If thread was successefully created
					if ( m_spMainThread.get() )
					{
						d_cerr( "ITIPServer::Start MainThread launched." );
						d_cerr( "" );

						m_bStarted = true;
						return true;
					}
				}
			}

			d_cerr( "" );
			return false;
		}

		// Stop main dispatcher thread
		virtual bool Stop()
		{
			d_cerr( "ITIPServer::Stop called." );
			d_cerr_sl( "m_bStarted: " ); d_cerr( m_bStarted ); 

			if ( m_bStarted )
			{
				// Temporary lock
				SrvLock TmpLock( m_mStarted );

				// Double checking with lock
				if ( m_bStarted )
				{
					// Set stop flag for main thread
					m_bHaveToStop = true;

					TmpLock.unlock();

					// Wait until main thread stop
					m_spMainThread -> join();
					m_bStarted = false;

					d_cerr( "ITIPServer::Stop MainThread stoped." );
					d_cerr( "" );

					return true;
				}
			}

			d_cerr( "" );
			return false;
		}

		// Add new portprocessor
		virtual bool AddProcessor( unsigned int port, IPortProcessor* proc, 
								   EDomain domain = D_INET, int options = 0 )
		{
			d_cerr( "ITIPServer::AddProcessor called." );
			d_cerr_sl( "Port: " ); d_cerr( port );
			d_cerr_sl( "Domain: " ); d_cerr( domain );
			d_cerr( "" );

			THREAD_DATA tdTmp;

			// Temporary lock
			SrvLock TmpLock( m_mContainer );

			// Copy passed values into the fields
			tdTmp.m_pPortProcessor = proc;

			tdTmp.m_EDomain = domain;
			tdTmp.m_iOptions = options;

			return m_mapThreads.insert( std::make_pair( port, tdTmp ) ).second;
		}

		// Delete portprocessor
		virtual bool DelProcessor( unsigned int port )
		{
			d_cerr( "ITIPServer::DelProcessor called." );
			d_cerr_sl( "Port: " ); d_cerr( port );
			d_cerr( "" );

			// Temporary lock
			SrvLock TmpLock( m_mContainer );

			// If there is nothing to delete
			if ( m_mapThreads.empty() )
				return false;

			// Try to find demanded port in container
			ThreadMap::iterator itPort = m_mapThreads.find( port );

			// If port was found
			if ( itPort != m_mapThreads.end() )
			{
				d_cerr_sl( "Record for port " ); d_cerr_sl( port ); 
				d_cerr( " found and will be removed." );
				d_cerr( "" );

				// Temporary lock
				SrvLock TmpStartLock( m_mStarted );

				if ( m_bStarted )
					// Set stop flag
					itPort -> second.m_bHaveToDel = true;
				else
					// Remove this record
					m_mapThreads.erase( itPort );

				return true;
			}

			return false;
		}

#ifdef USE_SSL
		void initSSL( SPTR_CSSL_CTX a_ptrCSSL_CTX )
		{
			m_ptrCSSL_CTX = a_ptrCSSL_CTX;
			isSSLused = true;
		}
#endif

	};

} // namespace IPStream

#endif // _IP_SERVER_H
