/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// Created by: Sergey Belovolov                                                         
// Created on: 26.05.04                                                                 
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

#include "IPSocket.h"
#include "boost_adopt/boost_adopt.h"

#ifdef __linux
	#include <errno.h>
	#include <signal.h>
#endif // __linux

#define TRACE_SOCKET_PROPERTIES( Point ) d_cerr( Point ); \
		d_cerr_sl( "Port: " ); d_cerr( m_usPortNumber ); \
		d_cerr_sl( "Protocol: " ); d_cerr( m_tpProtocol ); \
		d_cerr_sl( "Options: " ); d_cerr( m_iOptions ); \
		d_cerr_sl( "Domain: " ); d_cerr( m_dDomain ); \
		d_cerr_sl( "Socket: " ); d_cerr( m_sSocket ); \
		d_cerr_sl( "Address: " ); d_cerr( m_strAddress ); \
		d_cerr_sl( "good(): " ); d_cerr( good() )

namespace IPStream
{
// Substitution for type of protocol verification
#define PR_MISMATCH( PR ) ( !( good() && ( m_tpProtocol == PR ) ) )

// Size of sockaddr(sockaddr_in) structure
	static const unsigned int scuiSISize = sizeof( sockaddr );

// Lock for atomic operations with socket
//	typedef boost::recursive_mutex::scoped_lock MyLock;

#define IPLOCK BRMutex::scoped_lock slLock( m_mtxAtomic )

	// Exceptions
	except_ipsocket_logic_error::except_ipsocket_logic_error( const std::string& message ) :
	std::logic_error( message ) 
	{
	}

	except_e_pointer::except_e_pointer( const std::string& message ) :
	std::runtime_error( message ) 
	{
	}

	socket_exception::socket_exception( const std::string& message ) :
	std::logic_error( message ) 
	{
	}

	invalid_socket::invalid_socket( const std::string& message ) :
	socket_exception( message ) 
	{
	}

	transmission_failed::transmission_failed( const std::string& message ) :
	socket_exception( message ) 
	{
	}

	CIPSocket::SOCKET_PROPERTIES::FOR_SELECT::FOR_SELECT( long sec ) 
	{
		m_timeout.tv_sec = sec;
		m_timeout.tv_usec = 0;
	}

    void CIPSocket::SOCKET_PROPERTIES::SOCKET_PROPERTIES_COMMON( unsigned short PortNumber, 
                                                     const std::string& address,
                                                     ETransportProtocol protocol,
                                                     int options, 
                                                     EDomain domain, 
                                                     SOCKET sk
                                                     )
    {
        m_bOpened = false; 
        m_dDomain = ( domain == D_UNSUP ) ? D_INET : domain;
        m_usPortNumber = PortNumber; 
        m_tpProtocol = ( protocol == TP_RAW ) ? TP_TCP : protocol;
        m_iOptions = options;
        m_sSocket = sk ? sk : ( socket( m_dDomain, m_tpProtocol, m_iOptions ) );
        m_bIsConnected = false;
        m_iLastError = 0;
        m_strAddress = address;

        TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES Constructor." );
        d_cerr_sl( "SOCKET: " ); d_cerr( sk ); 

        if ( IsValid() )
        {
            d_cerr( "Socket is opened now!" );
            m_bOpened = true;
        }

        d_cerr( "" );

        // Pump the constructor data into the internal structures
        Fill();

        // Non-blocking - default mode
        SetNonBlocking();

#ifdef _TCP_NO_WAIT
        unsigned int uiOn = 1;

        // Disable the Nagle algorithm for send coalescing
        SetSockOpt( ( const char* ) &uiOn, sizeof( uiOn ), IPPROTO_TCP, TCP_NODELAY );
#endif // #ifdef _TCP_NO_WAIT

#ifdef _SPLIT_OUT_BUF
        int iOn = _SPLIT_OUT_BUF;

        // Set receive buffer size
        SetSockOpt( ( const char* ) &iOn, sizeof( iOn ), SOL_SOCKET, SO_RCVBUF );
        // Set send buffer size
        SetSockOpt( ( const char* ) &iOn, sizeof( iOn ), SOL_SOCKET, SO_SNDBUF );

#endif // _SPLIT_OUT_BUF
    }


	// Default constructor
	// N.B.: You'll have to check m_sSocket == INVALID_SOCKET (for windows at least)
	CIPSocket::SOCKET_PROPERTIES::SOCKET_PROPERTIES( unsigned short PortNumber, 
													 const std::string& address,
													 ETransportProtocol protocol,
													 int options, 
													 EDomain domain, 
													 SOCKET sk
                                                     )
	{
        SOCKET_PROPERTIES_COMMON( PortNumber, 
                                  address, 
                                  protocol, 
                                  options, 
                                  domain, 
                                  sk );
	}

#ifdef USE_SSL
    CIPSocket::SOCKET_PROPERTIES::SOCKET_PROPERTIES( SPTR_CSSL_CTX a_ptrCTX,
                       unsigned short PortNumber, 
                       const std::string& address,
                       int options, 
                       EDomain domain, 
                       SOCKET sk
                       )
    {
        SOCKET_PROPERTIES_COMMON(
              PortNumber, 
              address,
              TP_TCP,
              options,
              domain,
              sk );


        m_ssl.ptrCTX = a_ptrCTX;
        if ( m_ssl.ptrCTX )
        {
            m_ssl.isSSLused = true;
        }
    }
#endif



	// Destructor that closes SOCKET if its open
	CIPSocket::SOCKET_PROPERTIES::~SOCKET_PROPERTIES()
	{
		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES Destructor." );
		d_cerr( "" );

		// If socket opened then close it
		Close();
	}

	void CIPSocket::SOCKET_PROPERTIES::Fill()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Fill." );

		// Nullify our storage memory 
		memset( &m_sStorage, 0, scuiSISize );
		memset( &m_sRcvStorage, 0, scuiSISize );

		switch ( m_dDomain )
		{
			// Internet domain socket
			case D_INET:
				{
					m_sinStorage.sin_family = m_dDomain;

					if ( m_strAddress.empty() )
						// Address wasn't set
						m_sinStorage.sin_addr.s_addr = htonl( INADDR_ANY );
					else
					{
						hostent* hepTmp = gethostbyname( m_strAddress.c_str() );

						// If no errors
						if ( hepTmp )
							// Fulfill mined hostname info
							memcpy( &m_sinStorage.sin_addr, hepTmp->h_addr, 
									hepTmp->h_length );
						else
						{
							OnLogicError( "SOCKET_PROPERTIES: gethostbyname failed." );
						}
					}

					// Set portnumber
					m_sinStorage.sin_port = htons( m_usPortNumber );

					break;
				}
				// UNIX domain socket
			case D_UNIX:
				{
					m_sStorage.sa_family = m_dDomain;

					if ( !m_strAddress.empty() )
						strcpy( m_sStorage.sa_data, m_strAddress.c_str() );
					else
					{
						OnLogicError( "SOCKET_PROPERTIES: address string is empty." );
					}

					break;
				}
				// Unsupported socket's domain
			default:
				{
					OnLogicError( "SOCKET_PROPERTIES: Wrong domain." );
				}
		};

		// For initial using sendto by client
		if ( m_tpProtocol == TP_UDP )
			m_sRcvStorage = m_sStorage;

		d_cerr( "" );
		//FillSelect();
	}

	void CIPSocket::SOCKET_PROPERTIES::FillSelect()
	{
		// Lock access
		IPLOCK;

		// Clean and initialize bindings
		FD_ZERO( &m_strctForSelect.m_readfds );
		FD_SET( m_sSocket, &m_strctForSelect.m_readfds );

		m_strctForSelect.m_writefds = m_strctForSelect.m_readfds;
		m_strctForSelect.m_exceptfds = m_strctForSelect.m_readfds;

		//m_strctForSelect.m_timeout.tv_sec = sec;
	}

	bool CIPSocket::SOCKET_PROPERTIES::IsValid() const
	{
		// Lock access
		IPLOCK;

		return( m_sSocket && ( m_sSocket != -1 ) );
	}

	bool CIPSocket::SOCKET_PROPERTIES::good() const
	{
		// Lock access
		IPLOCK;

		// Win32 variant
		return( IsValid() && m_bOpened );
	}

	void CIPSocket::SOCKET_PROPERTIES::SetError()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::SetError." );

#ifdef _WIN32 // Compile under MD
		m_iLastError = WSAGetLastError();
#elif defined( __linux ) // Compile under Linux
		m_iLastError = errno;
#else // Unsupported platform
	#error "Unsupported platform!"
#endif // _WIN32

		d_cerr_sl( "LastError: " ); d_cerr( m_iLastError );
		d_cerr( "" );
	}

	bool CIPSocket::SOCKET_PROPERTIES::WasError()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::WasError." );

#ifdef _WIN32 // Compile under MD
		WSASetLastError( 0 );
#elif defined( __linux ) // Compile under Linux
		int iErr;
		int iLen = sizeof( iErr );

		errno = 0;
#else // Unsupported platform
	#error "Unsupported platform!"
#endif // _WIN32

#ifdef _WIN32 // Compile under MD
		if ( !FD_ISSET( m_sSocket, &m_strctForSelect.m_exceptfds ) )
			return false;

		SetError();

#elif defined( __linux ) // Compile under Linux
		if ( getsockopt( m_sSocket, SOL_SOCKET, SO_ERROR, &iErr, ( socklen_t* ) &iLen ) < 0 )
		{
			// If getsockopt failed
			SetError();
			return true;
		}

		// Check if no error happened
		if ( !iErr )
			return false;

		errno = iErr;
        
#else // Unsupported platform
	#error "Unsupported platform!"
#endif // _WIN32

		// Close current socket
		Close();

		return true;
	}

	int CIPSocket::SOCKET_PROPERTIES::GetLastError() const throw()
	{
		// Lock access
		IPLOCK;

		return m_iLastError;
	}

	// Call this after error to reveal if connection may be used further
	bool CIPSocket::SOCKET_PROPERTIES::IsBigMessage()
	{
		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::IsBigMessage." );
		d_cerr( "" );

		SetError();

#ifdef _WIN32
		switch ( m_iLastError )
		{
			// Message size is greater then the socket buffer (after recv)
			// Watch out - message was truncated
			case WSAEMSGSIZE:
				return true;
				break;
			default:
				break;
		}
#endif // _WIN32

		return false;
	}

	// Close SOCKET
	bool CIPSocket::SOCKET_PROPERTIES::Close()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Close." );
		d_cerr( "" );

		// Drop UDP-connected flag
		m_bIsConnected = false;

#ifdef USE_SSL
        if ( m_ssl.isSSLused ) {
            m_ssl.ssl.clean();
        }
#endif
		// Clean state 
		FD_ZERO( &m_strctForSelect.m_readfds );

		// If socket opened then close it
		if ( good() )
		{
			m_bOpened = false;
			return( ( -1 != closeskt( m_sSocket ) ) ? true : ( SetError(), false ) );
		}
		return false;
	}

	// Open SOCKET after calling Set or Close
	bool CIPSocket::SOCKET_PROPERTIES::Open()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Open." );

		Close();

		m_sSocket = socket( m_dDomain, m_tpProtocol, m_iOptions );

		d_cerr_sl( "Reopened socket: " ); d_cerr( m_sSocket ); d_cerr( "" );

		if ( IsValid() )
			return( m_bOpened = true );
		else
		{
			SetError();
			return false;
		}

        return false;
	}

	bool CIPSocket::SOCKET_PROPERTIES::IOCtl( long Cmd, unsigned long* arg )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::IOCtl." );
		d_cerr_sl( "Command: " ); d_cerr( Cmd );
		d_cerr_sl( "Argument: " ); d_cerr( *arg );
		d_cerr( "" );

		// If m_sSocket isn't acceptable
		if ( !good() || arg )
			return false;

		return( ioctl( m_sSocket, Cmd, arg ) ? ( SetError(), false ) : true );
	}

	// Socket management
	// By default - switches socket into reuse-addr mode 
	bool CIPSocket::SOCKET_PROPERTIES::SetSockOpt( const char* optval, int optlen,
												   int level, int optname )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::SetSockOpt." );
		d_cerr_sl( "optlen: " ); d_cerr( optlen );
		d_cerr_sl( "level: " ); d_cerr( level );
		d_cerr_sl( "optname: " ); d_cerr( optname );
		d_cerr( "" );

		if ( !( optval && good() ) )
			return false;

		return( setsockopt( m_sSocket, level, optname, optval, optlen ) ? 
				( SetError(), false ) : true );
	}

	void CIPSocket::SOCKET_PROPERTIES::SetNonBlocking( long sectimeout )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::SetNonBlocking." );
		d_cerr_sl( "sectimeout: " ); d_cerr( sectimeout );
		d_cerr( "" );

		if ( ( sectimeout < 1 ) || !good() )
			return;

		m_strctForSelect.m_timeout.tv_sec = sectimeout;
		m_strctForSelect.m_timeout.tv_usec = 0;

		unsigned long uiTmp = 1;

		// Set non-blocking
		IOCtl( FIONBIO, &uiTmp );
	}

	void CIPSocket::SOCKET_PROPERTIES::SetBlocking()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::SetBlocking." );
		d_cerr( "" );

		if ( good() )
		{
			// Select will be called in the blocking mode
			m_strctForSelect.m_timeout.tv_sec = 0;
			m_strctForSelect.m_timeout.tv_usec = 0;

			unsigned long uiTmp = 0;

			// Switch stream into blocking mode
			IOCtl( FIONBIO, &uiTmp ); 
		}
	}

	// Make timeout for socket
	bool CIPSocket::SOCKET_PROPERTIES::Select()
	{
		// Lock access
		IPLOCK;

		if ( !good() )
			return false;

		FillSelect();

        // Save previously stored timeout value (for Linux realization)
		unsigned int uiPrevTimeout = m_strctForSelect.m_timeout.tv_sec;

		/*
		// For temporary usage only. Without this TCPServer doesn't work in Linux.
		if ( !m_strctForSelect.m_timeout.tv_sec )
			m_strctForSelect.m_timeout.tv_sec = cuiSelectTimeout;
		*/

		int iSelRes = select( m_sSocket + 1, &m_strctForSelect.m_readfds, 
							  &m_strctForSelect.m_writefds, &m_strctForSelect.m_exceptfds,
							  ( m_strctForSelect.m_timeout.tv_sec ? 
								&m_strctForSelect.m_timeout : 0 ) ); 

		// Restore previously saved value
		if ( uiPrevTimeout )
			m_strctForSelect.m_timeout.tv_sec = uiPrevTimeout;

		return ( ( iSelRes < 0 ) ? ( SetError(), false ) : true );
	}

	bool CIPSocket::SOCKET_PROPERTIES::IsConnected()// const
	{
		// Lock access
		IPLOCK;

		if ( !good() )
			return false;

		if ( WasError() )
			return false;

		if ( !( FD_ISSET( m_sSocket, &m_strctForSelect.m_readfds ) || 
				FD_ISSET( m_sSocket, &m_strctForSelect.m_writefds ) ) )
			return false;

		return true;
	}

	bool CIPSocket::SOCKET_PROPERTIES::CanRead()// const
	{
		// Lock access
		IPLOCK;

		if ( !good() )
			return false;

		if ( WasError() )
			return false;

#ifdef USE_SSL
        if ( m_ssl.isSSLused && m_ssl.ssl.isChanceToRead() )
            return true;
#endif


		if ( FD_ISSET( m_sSocket, &m_strctForSelect.m_readfds ) )
			return true;

		return false;
	}

	// Check if we can write
	bool CIPSocket::SOCKET_PROPERTIES::CanWrite() //const
	{
		// Lock access
		IPLOCK;

		if ( !good() )
			return false;

		if ( WasError() )
			return false;

		if ( FD_ISSET( m_sSocket, &m_strctForSelect.m_writefds ) )
			return true;

		return false;
	}

	// Set the portnumber and address (for client only)
	// Watch out! This method reopens socket
	bool CIPSocket::SOCKET_PROPERTIES::Set( unsigned short port, 
											const std::string& address,
											ETransportProtocol protocol,
											int options,
											EDomain domain )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Set." );

		d_cerr_sl( "New port: " ); d_cerr( port );
		d_cerr_sl( "New protocol: " ); d_cerr( protocol );
		d_cerr_sl( "New options: " ); d_cerr( options );
		d_cerr_sl( "New domain: " ); d_cerr( domain );
		d_cerr_sl( "New address: " ); d_cerr( address ); 
		d_cerr( "" );

		m_iLastError = 0;

		m_usPortNumber = port;

		m_iOptions = options;

		if ( !address.empty() )
			m_strAddress = address;

		if ( protocol != TP_RAW )
			m_tpProtocol = protocol;

		if ( domain != D_UNSUP )
			m_dDomain = domain;

		if ( !Open() )
			return false;

		// Set data
		Fill();

		return true;
	}

	// Get values for socket's port and address
	void CIPSocket::SOCKET_PROPERTIES::Get( unsigned short* port, 
											std::string* address,
											ETransportProtocol* protocol,
											int* options,
											EDomain* domain ) const
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Get." );
		d_cerr( "" );

		// If nothing we can do
		if ( port )
			*port = m_usPortNumber;

		if ( address )
			*address = m_strAddress;

		if ( protocol )
			*protocol = m_tpProtocol;

		if ( options )
			*options = m_iOptions;

		if ( domain )
			*domain = m_dDomain;
	}

	bool CIPSocket::SOCKET_PROPERTIES::Bind()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Bind." );
		d_cerr( "" );

		// If m_sSocket isn't acceptable
		if ( !good() )
			return false;

		return( bind( m_sSocket, &m_sStorage, scuiSISize ) ? ( SetError(), false ) : true );
	}

	bool CIPSocket::SOCKET_PROPERTIES::Listen( int backlog )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Listen." );

		d_cerr_sl( "backlog: " ); d_cerr( backlog );
		d_cerr( "" );

		// If m_sSocket isn't acceptable
		if ( PR_MISMATCH( TP_TCP ) )
			return false;

		return( listen( m_sSocket, backlog ) ? ( SetError(), false ) : true );
	}

	// Watch out! It may return pointer to NULL
	CIPSocket::SOCKET_PROPERTIES::SPSmartPtr CIPSocket::SOCKET_PROPERTIES::Accept()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Accept." );

		// If m_sSocket isn't acceptable 
		if ( PR_MISMATCH( TP_TCP ) || !( Select() && IsConnected() ) )
		{
			d_cerr( "SOCKET_PROPERTIES::Accept - Somethig went wrong... accept cannot be successfully executed at this momemt." );
			d_cerr( "" );
			// Return NULL pointer
			return CIPSocket::SOCKET_PROPERTIES::SPSmartPtr();
		}

		int iTmp = scuiSISize;

		SOCKET sTmp = accept( m_sSocket, &m_sRcvStorage, 
// Explicit type convertion for Linux
#ifdef __linux          
							  ( socklen_t* ) 
#endif // __linux           
							  &iTmp );

		d_cerr_sl( "Accepted socket: " ); d_cerr( sTmp );
		d_cerr( "" );

		// Check result of the Accept
		if ( !sTmp || ( sTmp == -1 ) )
		{
			SetError();

			// Return NULL pointer
			return CIPSocket::SOCKET_PROPERTIES::SPSmartPtr();
		}

        CIPSocket::SOCKET_PROPERTIES::SPSmartPtr retval;

#ifdef USE_SSL
        if ( m_ssl.isSSLused && m_ssl.ptrCTX ) {

            retval = CIPSocket::SOCKET_PROPERTIES::SPSmartPtr(
                   new CIPSocket::SOCKET_PROPERTIES( 
                         m_ssl.ptrCTX,
                         m_usPortNumber, 
                         m_strAddress,
                         m_iOptions, 
                         m_dDomain, 
                         sTmp ) );
        } else
        {
            //compiled with SSL, but not used SSL structures
            retval = CIPSocket::SOCKET_PROPERTIES::SPSmartPtr(
                   new CIPSocket::SOCKET_PROPERTIES( m_usPortNumber, 
                         m_strAddress,
                         m_tpProtocol,
                         m_iOptions, 
                         m_dDomain, 
                         sTmp ) );
        }

#else
		// Create new smartpointer and return it
        retval = CIPSocket::SOCKET_PROPERTIES::SPSmartPtr(
               new CIPSocket::SOCKET_PROPERTIES( m_usPortNumber, 
                     m_strAddress,
                     m_tpProtocol,
                     m_iOptions, 
                     m_dDomain, 
                     sTmp ) );
#endif



#ifdef USE_SSL
        int iSSLblockingLoop =0;
        for (iSSLblockingLoop=0; iSSLblockingLoop<5; iSSLblockingLoop++) {
            if ( retval->good() ) break;
            boost_adopt::sleep(1);
        }

        d_cerr_sl("SSL socket accept blocking loop, cycles ");
        d_cerr(iSSLblockingLoop);

        if ( iSSLblockingLoop == 5) {
            d_cerr("SSL socket accept blocking loop failed");
            return retval;
        }


        if ( retval && retval->m_ssl.isSSLused && retval->m_ssl.ptrCTX ) {
            if ( retval->m_ssl.ssl.SSL_accept( retval->m_ssl.ptrCTX, sTmp ) )
            {
                d_cerr("SSL m_ssl.ssl.SSL_accept server completed" );
            } else
            {
                d_cerr("SSL m_ssl.ssl.SSL_accept server failed" );
            };
        };
#endif


        return retval;
	}

	// false - error occured
	bool CIPSocket::SOCKET_PROPERTIES::Connect()
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Connect." );
		d_cerr( "" );

		// If m_sSocket isn't acceptable
		if ( !good() )
			return false;

		bool bTmp = connect( m_sSocket, ( ( m_tpProtocol == TP_UDP ) ? &m_sRcvStorage : &m_sStorage ), 
							 scuiSISize ) ? ( SetError(), false ) : true;

		Select();

		// If we want to get connected UDP-socket
		if ( m_tpProtocol == TP_UDP )
			m_bIsConnected = bTmp;

		//return bTmp;
		bool retval = IsConnected();

#ifdef USE_SSL
        if ( retval && m_ssl.isSSLused && m_ssl.ptrCTX ) {
            if ( m_ssl.ssl.SSL_connect( m_ssl.ptrCTX, m_sSocket ) )
            {
                d_cerr("SSL m_ssl.ssl.SSL_connect client completed" );
            } else
            {
                d_cerr("SSL m_ssl.ssl.SSL_connect client failed" );
                retval = false;
            };
        };
#endif

        return retval;

	}

	int CIPSocket::SOCKET_PROPERTIES::Send( const char* outbuf, int len, int flag )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Send." );
		d_cerr_sl( "len: " ); d_cerr( len );
		d_cerr_sl( "flag: " ); d_cerr( flag );
		d_cerr( "" );

		// If m_sSocket isn't acceptable or data not suitable
		if ( !( good() && outbuf && len ) )
			return -1;

		int iTmp = 0;

		switch ( m_tpProtocol )
		{
			case TP_TCP:
				{
					if ( Select() && CanWrite() )
					{
#ifdef _SPLIT_OUT_BUF
						while ( len > _SPLIT_OUT_BUF )
						{
#ifdef USE_SSL
                            if ( m_ssl.isSSLused && m_ssl.ssl.good() ) {
                                iTmp = m_ssl.ssl.SSL_write( outbuf, _SPLIT_OUT_BUF );
                            } else
#endif
                            {
#ifdef USE_SSL
                                d_cerr( "SSL_write: error m_ssl.ssl.good() false" );
#endif
                                iTmp = send( m_sSocket, outbuf, _SPLIT_OUT_BUF, flag );
                            }
							outbuf += _SPLIT_OUT_BUF;
							len -= _SPLIT_OUT_BUF;
						}
#endif // _SPLIT_OUT_BUF
						// Send the residual part of buffer
						if ( len )
                        {
#ifdef USE_SSL
                            if ( m_ssl.isSSLused && m_ssl.ssl.good() ) {
                                iTmp = m_ssl.ssl.SSL_write( outbuf, len );
                            } else

#endif  
                            {
#ifdef USE_SSL
                                d_cerr( "SSL_write: error m_ssl.ssl.good() false" );
#endif
                           
                                iTmp = send( m_sSocket, outbuf, len, flag );
                            }
                        };

						d_cerr_sl( "send result: " ); d_cerr( iTmp );
						d_cerr( "" );

						if ( iTmp </*=*/ 0 )
							SetError();

						return iTmp;
					} // if ( Select() && CanWrite() )

					return 0;
					break;
				}
			case TP_UDP:
				{
					if ( Select() && CanWrite() )
					{
						d_cerr_sl( "m_bIsConnected: " ); d_cerr( m_bIsConnected );

						// If our UDP-socket is connected
						if ( m_bIsConnected )
						{
#ifdef _SPLIT_OUT_BUF
							while ( len > _SPLIT_OUT_BUF )
							{
								iTmp = send( m_sSocket, outbuf, _SPLIT_OUT_BUF, flag );

								outbuf += _SPLIT_OUT_BUF;
								len -= _SPLIT_OUT_BUF;
							}
#endif // _SPLIT_OUT_BUF
							// Send the residual part of buffer
							if ( len )
								iTmp = send( m_sSocket, outbuf, len, flag );
						}
						else
						{
#ifdef _SPLIT_OUT_BUF
							while ( len > _SPLIT_OUT_BUF )
							{
								iTmp = sendto( m_sSocket, outbuf, _SPLIT_OUT_BUF,
											 flag, &m_sRcvStorage, scuiSISize );

								outbuf += _SPLIT_OUT_BUF;
								len -= _SPLIT_OUT_BUF;
							}
#endif // _SPLIT_OUT_BUF
							if ( len )
								iTmp = sendto( m_sSocket, outbuf, len, flag, 
										   &m_sRcvStorage, scuiSISize );
						}
							

						d_cerr_sl( "send result: " ); d_cerr( iTmp );
						d_cerr( "" );

						if ( iTmp </*=*/ 0 )
							SetError();

						return iTmp;

					} // if ( Select() && CanWrite() )

					return 0;
					break;
				}
			default:
				OnLogicError( "Send: Wrong protocol." );
		};

        return 0;
	}

	int CIPSocket::SOCKET_PROPERTIES::Recv( char* inbuf, int len, int flag )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::Recv." );
		d_cerr_sl( "len: " ); d_cerr( len );
		d_cerr_sl( "flag: " ); d_cerr( flag );
		d_cerr( "" );

		// If m_sSocket isn't acceptable or data not suitable
		if ( !( good() && inbuf && len ) )
			return -1;

		int iTmp = scuiSISize;

		switch ( m_tpProtocol )
		{
			case TP_TCP:
				{
					if ( Select() && CanRead() )
					{
#ifdef USE_SSL
                        if ( m_ssl.isSSLused && m_ssl.ssl.good() ) 
						{
                            iTmp = m_ssl.ssl.SSL_read( inbuf, len );
                        } 
						else
#endif
                        {
#ifdef USE_SSL
                            d_cerr( "SSL_read: error m_ssl.ssl.good() false" );
#endif
							// Unplug SIGPIPE rising on STREAM sockets
                            iTmp = recv( m_sSocket, inbuf, len, ( flag ? flag : 
#ifdef __linux           
							MSG_NOSIGNAL
#else
							0
#endif // __linux							
							 ) );
                        }

						d_cerr_sl( "recv result: " ); d_cerr( iTmp );
						d_cerr( "" );

						// Check if message size if greater then the buffer size
						if ( ( iTmp < 0 ) && IsBigMessage() )
							iTmp = len;

						return iTmp;

					} // if ( Select() && CanRead() )

					return ciHasNoData;
					break;
				}
			case TP_UDP:
				{

					if ( Select() && CanRead() )
					{

						d_cerr_sl( "m_bIsConnected: " ); d_cerr( m_bIsConnected );

						// If our UDP-socket is connected
						if ( m_bIsConnected )
							// Unplug SIGPIPE rising on STREAM sockets
							iTmp = recv( m_sSocket, inbuf, len, ( flag ? flag : 
#ifdef __linux           
							MSG_NOSIGNAL
#else
							0
#endif // __linux						
						) );
						
						else
							iTmp = recvfrom( m_sSocket, inbuf, len, flag, 
											 &m_sRcvStorage, 
#ifdef __linux           
											 ( socklen_t* ) 
#endif // __linux
											 &iTmp );

						d_cerr_sl( "recv result: " ); d_cerr( iTmp );
						d_cerr( "" );

						// Check if message size if greater then the buffer size
						if ( ( iTmp < 0 ) && IsBigMessage() )
							iTmp = len;

						return iTmp;

					} // if ( Select() && CanRead() )

					return ciHasNoData;
					break;
				}
			default:
				{
					OnLogicError( "Recv: Wrong protocol." );
				}
		};

        return 0;
	}

	bool CIPSocket::SOCKET_PROPERTIES::GetPeerInfo( std::string* address,
													unsigned short* port )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::GetPeerInfo." );
		d_cerr( "" );

		// If m_sSocket isn't acceptable or data not suitable
		if ( !address )
			return false;

		switch ( m_dDomain )
		{
			// Internet domain socket
			case D_INET:
				{
					if ( port )
					{
						*port = ntohs( m_sinRcvStorage.sin_port );

						d_cerr_sl( "port: " ); d_cerr( *port );
					}
                        
					unsigned long laddr;

					// If we can't get inet address
					if ( static_cast < int > 
						 ( laddr = inet_addr( inet_ntoa( m_sinRcvStorage.sin_addr ) ) ) == -1 )
					{
						SetError();
						return false;
					}

					hostent* hp = gethostbyaddr( ( char* ) &laddr, sizeof( laddr ), D_INET );

					// If we can't get name by address
					if ( !( hp && hp->h_name ) )
					{
						SetError();
						return false;
					}

					*address = hp->h_name;

					d_cerr_sl( "address: " ); d_cerr( *address );

					return true;

					break;
				}
				// UNIX domain socket
			case D_UNIX:
				{
					*address = m_sRcvStorage.sa_data;

					d_cerr_sl( "address: " ); d_cerr( *address );

					return true;

					break;
				}
				// Unsupported socket's domain
			default:
				{
					OnLogicError( "SOCKET_PROPERTIES: Wrong domain." );
				}
		};
        return false;
	}

	void CIPSocket::SOCKET_PROPERTIES::CopyPeerInfo( CIPSocket::SOCKET_PROPERTIES& dest )
	{
		// Lock access
		IPLOCK;

		TRACE_SOCKET_PROPERTIES( "SOCKET_PROPERTIES::CopyPeerInfo." );
		d_cerr( "" );

		// Copy peer info
		dest.m_sinRcvStorage = m_sinRcvStorage;
	}

	void CIPSocket::SOCKET_PROPERTIES::OnLogicError( const std::string& message )
	{
		SetError();

		d_cerr( message );
		throw except_ipsocket_logic_error( message );
	}

#if defined( __linux ) && !defined( USE_SSL ) // Compile under Linux
	void CIPSocket::sigPipeHandler( int )
	{
	}
#endif // defined( __linux ) && !defined( USE_SSL )

	void CIPSocket::OnRuntimeError( const std::string& message ) const
	{
		d_cerr( message );
		throw except_e_pointer( message );
	}



	CIPSocket::CIPSocket( unsigned short PortNumber,
						  const std::string& address,
						  ETransportProtocol protocol, 
						  int options, 
						  EDomain domain 
                          ) : 
	m_spSocket( new CIPSocket::SOCKET_PROPERTIES( PortNumber, 
												  address, 
												  protocol,
												  options, 
												  domain 
                                                  ) )
	{
#if defined( __linux ) && !defined( USE_SSL ) // Compile under Linux
		// Cut off SIGPIPE
        signal( SIGPIPE, sigPipeHandler );
#endif // defined( __linux ) && !defined( USE_SSL )
	}

#ifdef USE_SSL
    CIPSocket::CIPSocket( SPTR_CSSL_CTX a_ptrCTX,
                          unsigned short PortNumber,
                          const std::string& address,
                          int options, 
                          EDomain domain 
                          ) : 
    m_spSocket( new CIPSocket::SOCKET_PROPERTIES( a_ptrCTX,
                                                  PortNumber, 
                                                  address, 
                                                  options, 
                                                  domain 
                                                  ) )
    {
    }
#endif


	// We need this for effective work with Accept method
	CIPSocket::CIPSocket( const CIPSocket::SOCKET_PROPERTIES::SPSmartPtr& AliveSocket ) :
	m_spSocket( AliveSocket )
	{
	}

	CIPSocket::~CIPSocket()
	{
	}

	bool CIPSocket::IsValid() const
	{
		return( m_spSocket.get() != 0 );
	}

	bool CIPSocket::good() const
	{
		return( IsValid() && ( m_spSocket -> good() ) );
	}

	// Get last error code
	int CIPSocket::GetLastError() const throw()
	{
		if ( IsValid() )
			return m_spSocket -> GetLastError();

		OnRuntimeError( "Invalid pointer in GetLastError." );

        return 0;
	}

	bool CIPSocket::Close()
	{
		if ( IsValid() )
			return m_spSocket -> Close();

		OnRuntimeError( "Invalid pointer in Close." );
        return false;
	}

	// Open SOCKET after calling Set
	bool CIPSocket::Open()
	{
		if ( IsValid() )
			return m_spSocket -> Open();

		OnRuntimeError( "Invalid pointer in Close." );
        return false;
	}

	bool CIPSocket::IOCtl( long Cmd, unsigned long* arg )
	{
		if ( IsValid() )
			return m_spSocket -> IOCtl( Cmd, arg );

		OnRuntimeError( "Invalid pointer in Close." );
        return false;
	}

	// Socket management
	// By default - switches socket into reuse-addr mode 
	bool CIPSocket::SetSockOpt( const char* optval, int optlen,
								int level, int optname )
	{
		if ( IsValid() )
			return m_spSocket -> SetSockOpt( optval, optlen, level, optname );

		OnRuntimeError( "Invalid pointer in SetSockOpt." );
        return false;
	}

	// Switch socket to non-blocking mode
	void CIPSocket::SetNonBlocking( long sectimeout )
	{
		if ( IsValid() )
			return m_spSocket -> SetNonBlocking( sectimeout );

		OnRuntimeError( "Invalid pointer in SetNonBlocking." );
	}

	// Switch socket to blocking mode
	void CIPSocket::SetBlocking()
	{
		if ( IsValid() )
			return m_spSocket -> SetBlocking();

		OnRuntimeError( "Invalid pointer in SetBlocking." );
	}

	bool CIPSocket::Select()
	{
		if ( IsValid() && good() )
			return m_spSocket -> Select();

		OnRuntimeError( "Invalid pointer or bad socket in Select." );
        return false;
	}

	bool CIPSocket::IsConnected() const
	{
		if ( IsValid() )
			return m_spSocket -> IsConnected();

		OnRuntimeError( "Invalid pointer in IsConnected." );
        return false;
	}

	// Check if we can read
	bool CIPSocket::CanRead() const
	{
		if ( IsValid() )
			return m_spSocket -> CanRead();

		OnRuntimeError( "Invalid pointer in CanRead." );
        return false;
	}

	// Check if we can write
	bool CIPSocket::CanWrite() const
	{
		if ( IsValid() )
			return m_spSocket -> CanWrite();

		OnRuntimeError( "Invalid pointer in CanWrite." );
        return false;
	}

	// Set the portnumber and address (for client only)
	bool CIPSocket::Set( unsigned short port, 
						 const std::string& address, 
						 ETransportProtocol protocol,
						 int options,
						 EDomain domain )
	{
		if ( IsValid() )
			return m_spSocket -> Set( port, address, protocol, options, domain );

		OnRuntimeError( "Invalid pointer in Set." );
        return false;
	}

	// Get values for socket's port and address
	void CIPSocket::Get( unsigned short* port, 
						 std::string* address,
						 ETransportProtocol* protocol,
						 int* options,
						 EDomain* domain ) const
	{
		if ( IsValid() )
			m_spSocket -> Get( port, address, protocol, options, domain );
		else
			OnRuntimeError( "Invalid pointer in Get." );
	}

	// false - error occured
	bool CIPSocket::Bind()
	{
		if ( good() )
			return m_spSocket -> Bind();

		OnRuntimeError( "Invalid pointer or bad socket in Bind." );
        return false;
	}

	// false - error occured
	bool CIPSocket::Listen( int backlog )
	{
		if ( good() )
			return m_spSocket -> Listen( backlog );

		OnRuntimeError( "Invalid pointer or bad socket in Listen." );
        return false;
	}

	CIPSocket CIPSocket::Accept()
	{
		if ( good() )
			return CIPSocket( m_spSocket -> Accept() );

		OnRuntimeError( "Invalid pointer or bad socket in Accept." );
        return CIPSocket();
	}

	bool CIPSocket::Connect()
	{
		if ( good() )
			return m_spSocket -> Connect();

		OnRuntimeError( "Invalid pointer or bad socket in Connect." );
        return false;
	}

	int CIPSocket::Send( const char* outbuf, int len, int flag )
	{
		if ( good() )
			return m_spSocket -> Send( outbuf, len, flag );

		OnRuntimeError( "Invalid pointer or bad socket in Send." );
        return 0;
	}

	int CIPSocket::Recv( char* inbuf, int len, int flag )
	{
		if ( good() )
			return m_spSocket -> Recv( inbuf, len, flag );

		OnRuntimeError( "Invalid pointer or bad socket in Recv." );
        return 0;
	}

	bool CIPSocket::GetPeerInfo( std::string* address, 
								 unsigned short* port ) const
	{
		if ( IsValid() )
			return m_spSocket -> GetPeerInfo( address, port );

		OnRuntimeError( "Invalid pointer in GetPeerInfo." );
        return false;
	}

	void CIPSocket::CopyPeerInfo( CIPSocket& dest )
	{
		if ( IsValid() )
			return m_spSocket -> CopyPeerInfo( *dest.m_spSocket );

		OnRuntimeError( "Invalid pointer in CopyPeerInfo." );
	}

} // namespace IPStream

