/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// Created by: Sergey Belovolov                                                         
// Created on: 26.05.04                                                                 
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

// Do not include directly

#ifndef _IPSOCKET_H
#define _IPSOCKET_H

#include "definitions.h"

#include <string>
#include <stdexcept>

#include "boost/shared_ptr.hpp"
#include "boost/thread/recursive_mutex.hpp"
#include "boost/thread/thread.hpp"
            
namespace IPStream
{
	// Timeout in seconds
	const unsigned int cuiSelectTimeout = 5;

	// If no data was received
	const int ciHasNoData = -2;

	// Protocols enumeration
	enum ETransportProtocol
	{
		// Raw socket (unsupported)
		TP_RAW = SOCK_RAW,
		// UDP
		TP_UDP = SOCK_DGRAM,
		// TCP
		TP_TCP = SOCK_STREAM
	};

	// Allowed domains enumeration
	enum EDomain
	{
		// Unsupported domain
		D_UNSUP = AF_IPX,
		// Internet domain
		D_INET = AF_INET,
		// UNIX domain
		D_UNIX = AF_UNIX
	};

	// Exceptions
	class IPSOCKET_API except_ipsocket_logic_error : public std::logic_error
	{
	public:
		except_ipsocket_logic_error( const std::string& message = "IPSocket type mismatch." );
	};

	class IPSOCKET_API except_e_pointer : public std::runtime_error
	{
	public:
		except_e_pointer( const std::string& message = "Invalid pointer." );
	}; 

	// Base exception for signaling about IPStream events
	class IPSOCKET_API socket_exception : public std::logic_error
	{
	public:
		socket_exception( const std::string& message = "IPSocket: something went wrong." );
	};

	// If soket become invalid
	class IPSOCKET_API invalid_socket : public socket_exception
	{
	public:
		invalid_socket( const std::string& message = "IPSocket can't perform requested operation." );
	};

	// If data transmission failed
	class IPSOCKET_API transmission_failed : public socket_exception
	{
	public:
		transmission_failed( const std::string& message = "IPSocket data transmission failed." );
	};

/**
 * Types definitions section
 */
	// Local mutex type
	typedef boost::mutex BMutex;

	// Local recursive mutex type
	typedef boost::recursive_mutex BRMutex;

	// Smart-pointer to thread type
	typedef boost::shared_ptr < boost::thread > SPThread;
/**
 * End of section
 */
	// Wrapper for working with socket. Copyable
	class IPSOCKET_API CIPSocket
	{
	private:

		// Wrapper over SOCKET
		// Uncopyable due to the boost::recursive_mutex presence
		// Thread-safe due to the all public methods are atomic one
		struct SOCKET_PROPERTIES
		{
			private:
			// Mutex for atomic access to SOCKET (thread-safety)
			// Recursive lock for calling one atomic function from another one
			// This object type is uncopyable
			mutable BRMutex m_mtxAtomic;

			// Structure for using API select
			struct FOR_SELECT
			{
				fd_set m_readfds;
				fd_set m_writefds;
				fd_set m_exceptfds;
				timeval m_timeout;

				// Default timeout - 5 seconds
				FOR_SELECT( long sec = cuiSelectTimeout );
			};

#ifdef USE_SSL
            //structure to keep SSL data
            struct FOR_SSL
            {
                bool isSSLused;
                SPTR_CSSL_CTX ptrCTX;
                CSSL ssl;

                FOR_SSL() : isSSLused( false ) {};
            };
#endif

			// Fill the internal structures
			void Fill();

			// Fill structure for Select call
			void FillSelect();

			// Action on logic error
			void OnLogicError( const std::string& message = "IPSocket type mismatch." );

			// Method for setting up error id
			void SetError();

			// Check if any errors with socket happened
			bool WasError();


			public:
			// Type for convenience of further usage
			typedef boost::shared_ptr < SOCKET_PROPERTIES > SPSmartPtr;

			// If SOCKET opened
			bool m_bOpened;

			// AF_UNIX (available in Linux) or AF_INET
			EDomain m_dDomain;

			// Port number for connection
			unsigned short m_usPortNumber;

			// Protocol that socket uses
			ETransportProtocol m_tpProtocol;

			// Socket's options
			int m_iOptions;

			// Socket for connection
			SOCKET m_sSocket;

			// For UDP sockets only
			// Indicates if this socket is UDP-connected
			bool m_bIsConnected;

			// The id-number of the last occured error
			int m_iLastError;

			// sockaddr_in for AF_INET
			// sockaddr for AF_UNIX
			union
			{
				// Storage for working with Internet Domain sockets 
				sockaddr_in m_sinStorage;

				// Storage for working with UNIX Domain sockets
				sockaddr m_sStorage;
			};

			// sockaddr_in for AF_INET
			// sockaddr for AF_UNIX
			// This union is for storing info about peer connection
			// which obtained from accept and recvfrom APIs
			union
			{
				// Storage for working with Internet Domain sockets 
				sockaddr_in m_sinRcvStorage;

				// Storage for working with UNIX Domain sockets
				sockaddr m_sRcvStorage;
			};


			// Address (for client socket only)
			std::string m_strAddress;

			// Settings for select call
			FOR_SELECT m_strctForSelect;

#ifdef USE_SSL
            FOR_SSL m_ssl;
#endif

			// Get last error code
			int GetLastError() const throw();

			// Call this after error to reveal if connection may be used further
			bool IsBigMessage();

			// Indicates that socket is valid
			bool IsValid() const;

////////////////////////////////////////////////////////////////
// Common section

			// Check if socket can be used
			bool good() const;

			// Close SOCKET
			bool Close();

			// Open SOCKET after calling Set
			bool Open();

			// Socket management
			// By default - switches socket into non-blocking mode
			bool IOCtl( long Cmd = FIONBIO, unsigned long* arg = 0 );

			// Socket management
			// By default - switches socket into reuse-addr mode 
			bool SetSockOpt( const char* optval = 0, int optlen = 0,
							 int level = SOL_SOCKET, int optname = SO_REUSEADDR );

			// Switch socket to non-blocking mode
			void SetNonBlocking( long sectimeout = cuiSelectTimeout );

			// Switch socket to blocking mode
			void SetBlocking();

			// Make timeout for socket
			bool Select();

			// Method for calling after connect call for checking if 
			// connection established
			bool IsConnected(); // const;

			// Check if we can read
			bool CanRead(); // const;

			// Check if we can write
			bool CanWrite(); // const;

			// Set the portnumber and address (for client only)
			bool Set( unsigned short port, 
					  const std::string& address = "", 
					  ETransportProtocol protocol = TP_RAW,
					  int options = 0, 
					  EDomain domain = D_UNSUP );

			// Get values for socket's port and address
			void Get( unsigned short* port, 
					  std::string* address = 0,
					  ETransportProtocol* protocol = 0,
					  int* options = 0, 
					  EDomain* domain = 0 ) const;

			// false - error occured
			bool Bind();

////////////////////////////////////////////////////////////////
// TCP section

			// false - error occured
			bool Listen( int backlog = SOMAXCONN );

			// Allocates new SOCKET_PROPERTIES via calling new operator
			// Result must be assigned to CIPSocket::m_spSocket in the following way:
			// {
			//  CIPSocket ipsTemp;
			//  ipsTemp.m_spSocket = m_spSocket -> Accept();    
			// }
			SPSmartPtr Accept();

			// false - error occured
			bool Connect();

			// Send TCP data
			// -1 - fail, else number of transferred bytes
			int Send( const char* outbuf, int len, int flag = 0 );

			// Receive TCP data
			// -1 - error, 0 - socket closed, else number of received bytes
			int Recv( char* inbuf, int len, int flag = 0 );

			// Get information about the remote side of connection
			bool GetPeerInfo( std::string* address, 
							  unsigned short* port = 0 );

			// Copy peer info of current SOCKET_PROPERTIES to dest
			void CopyPeerInfo( SOCKET_PROPERTIES& dest );

            //common construction
            void SOCKET_PROPERTIES_COMMON( unsigned short PortNumber = 0, 
							   const std::string& address = "",
							   ETransportProtocol protocol = TP_TCP,
							   int options = 0, 
							   EDomain domain = D_INET, 
							   SOCKET sk = 0
                               );

			// For nullifying structure members
			SOCKET_PROPERTIES( unsigned short PortNumber = 0, 
							   const std::string& address = "",
							   ETransportProtocol protocol = TP_TCP,
							   int options = 0, 
							   EDomain domain = D_INET, 
							   SOCKET sk = 0
                               );
#ifdef USE_SSL
            /**
             * Constructor for SSL usage
             * @param a_ptrCTX smart pointer to CSSL_CTX object
             */
            SOCKET_PROPERTIES( SPTR_CSSL_CTX a_ptrCTX,
                               unsigned short PortNumber, 
                               const std::string& address,
                               //SSL is for TCP only
                               //ETransportProtocol protocol = TP_TCP,
                               int options = 0, 
                               EDomain domain = D_INET, 
                               SOCKET sk = 0
                               );
#endif


			// Destructor that closes SOCKET if its open
			~SOCKET_PROPERTIES(); 
		};

		// Smart-pointer to SOCKET_PROPERTIES object
		// It's behaviour is similar to a singletone object
		// Convert uncopyable SOCKET_PROPERTIES into copyable smart-pointer
		SOCKET_PROPERTIES::SPSmartPtr m_spSocket;

		// Copy constructor
		CIPSocket( const SOCKET_PROPERTIES::SPSmartPtr& );

		// Action on runtime error
		void OnRuntimeError( const std::string& message = "Invalid pointer." ) const;

#if defined( __linux ) && !defined( USE_SSL ) // Compile under Linux
		static void sigPipeHandler( int );
#endif // defined( __linux ) && !defined( USE_SSL )

	public:

		CIPSocket( unsigned short PortNumber = 0, 
				   const std::string& address = "",
				   ETransportProtocol protocol = TP_TCP, 
				   int options = 0, 
				   EDomain domain = D_INET
                   );

#ifdef USE_SSL
        CIPSocket( SPTR_CSSL_CTX a_ptrCTX,
                   unsigned short PortNumber,
                   const std::string& address,
                   int options = 0, 
                   EDomain domain = D_INET
                    );
#endif


		virtual ~CIPSocket();

		// Indicates that socket is valid
		bool IsValid() const;

		// Analog for method good of std streams
		bool good() const;

		// Get last error code
		int GetLastError() const throw();
////////////////////////////////////////////////////////////////
// Common section

		// Close opened socket
		bool Close();

		// Open SOCKET after calling Set
		bool Open();

		// Socket management
		// By default - switches socket into non-blocking mode
		bool IOCtl( long Cmd = FIONBIO, unsigned long* arg = 0 );

		// Socket management
		// By default - switches socket into reuse-addr mode 
		bool SetSockOpt( const char* optval = 0, int optlen = 0,
						 int level = SOL_SOCKET, int optname = SO_REUSEADDR );

		// Switch socket to non-blocking mode
		void SetNonBlocking( long sectimeout = cuiSelectTimeout );

		// Switch socket to blocking mode
		void SetBlocking();

		// Make timeout for socket
		bool Select();

		// Method for calling after connect call for checking if 
		// connection established
		bool IsConnected() const;

		// Check if we can read
		bool CanRead() const;

		// Check if we can write
		bool CanWrite() const;

		// Set the portnumber and address (for client only)
		bool Set( unsigned short port, 
				  const std::string& address = "",
				  ETransportProtocol protocol = TP_RAW,
				  int options = 0, 
				  EDomain domain = D_UNSUP );

		// Get values for socket's port and address
		void Get( unsigned short* port, 
				  std::string* address = 0,
				  ETransportProtocol* protocol = 0,
				  int* options = 0, 
				  EDomain* domain = 0 ) const;

		// false - error occured
		bool Bind();

////////////////////////////////////////////////////////////////
// TCP section

		// false - error occured
		bool Listen( int backlog = SOMAXCONN );

		// Create new SOCKET if accept() succeeded
		// If accept() failed then new socket will be uninitialized
		CIPSocket Accept();

		// false - error occured
		bool Connect();

		// Send TCP data
		// -1 - fail, else number of transferred bytes
		int Send( const char* outbuf, int len, int flag = 0 );

		// Receive TCP data
		// -1 - error, 0 - socket closed, else number of received bytes
		int Recv( char* inbuf, int len, int flag = 0 );

		// Get information about the remote side of connection (can be used only by server side)
		bool GetPeerInfo( std::string* address, 
						  unsigned short* port = 0 ) const;

		// Copy peer info of current CIPSocket to dest
		// It may be usefull when creating socket via accept call
		void CopyPeerInfo( CIPSocket& dest );
	};

} // namespace IPStream

#endif // _IPSOCKET_H

