/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// File: SSLcommon.cpp
// Author: Dmitry Ignatyuk
// Date: 16.10.04                                                                 
// Modified by:
// Last update:                                                                                     
// Copyright:                                                                                     
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

// Don't change order of following two files including
// That's important for compiling under Windows
#include "definitions.h"
#include "SSLcommon.h"

// Check if we really need to include implementation
#ifdef USE_SSL
	
	#include "boost_adopt/boost_adopt.h"

	#include <iostream>

#ifdef __linux
	#include <signal.h>
#endif // __linux

using namespace std;

CRYPTO_dynlock_value* CSSL_CTX::dyn_create_function( const char* file, int line )
{
	struct CRYPTO_dynlock_value* value;
	value = ( struct CRYPTO_dynlock_value* ) malloc( sizeof( struct CRYPTO_dynlock_value ) );
	if ( !value )
		return 0;

	MUTEX_SETUP( value -> mutex );
	return value;
}

void CSSL_CTX::locking_function( int mode, int n, const char *file, int line )
{
	if ( mode & CRYPTO_LOCK )
		MUTEX_LOCK( mutex_buf[ n ] );
	else
		MUTEX_UNLOCK( mutex_buf[ n ] );
}

unsigned long CSSL_CTX::id_function()
{
	return ( ( unsigned long ) THREAD_ID );
}

void CSSL_CTX::dyn_lock_function( int mode, struct CRYPTO_dynlock_value* l, const char* file, int line )
{
	if ( mode & CRYPTO_LOCK )
		MUTEX_LOCK( l -> mutex );
	else
		MUTEX_UNLOCK( l -> mutex );
}

void CSSL_CTX::dyn_destroy_function( struct CRYPTO_dynlock_value* l, const char* file, int line )
{
	MUTEX_CLEANUP( l -> mutex );
	free( l );
}

	#ifdef __linux
int CSSL_CTX::SSL_seed_prng( int bytes )
{
	if ( !RAND_load_file( "/dev/random", bytes ) )
		return 0;

	return 1;
}
	#else
int CSSL_CTX::SSL_seed_prng( int bytes )
{
	return 1;
}
	#endif // __linux


///////////////////////////////////////////////////////////////////////
//
//                 CSSL_CTX class implementation
//
///////////////////////////////////////////////////////////////////////

DH* CSSL_CTX::dh512 = 0;
DH* CSSL_CTX::dh1024 = 0;
MUTEX_TYPE* CSSL_CTX::mutex_buf = 0;


CSSL_CTX::CSSL_CTX()    
{
	ctx = 0;
}

CSSL_CTX::~CSSL_CTX()   
{
	if ( ctx )
	{
		SSL_CTX_free( ctx );
		ctx = 0;
	}
}

int CSSL_CTX::CTX_passwd_callback(
								 char* buf, 
								 int size, 
								 int flag, 
								 void* userdata )
{
	assert( userdata );
	if ( !userdata )
		return 0;

	int term = ( ( size - 1 ) < ( int ) strlen( ( const char* ) userdata ) ) ? 
		size - 1 : strlen( ( const char* ) userdata );
	strncpy( buf, ( const char* ) userdata, term );
	buf[ term  ] = 0;

	return term;
}

int CSSL_CTX::CTX_verify_callback(
								 int ok, 
								 X509_STORE_CTX* store )
{
	char data[ 256 ];
	char strBuf[ 1024 ];
	if ( !ok )
	{
		d_cerr( "ssl_CTX_verify_callback error:" );
		X509 *cert = X509_STORE_CTX_get_current_cert( store );
		int depth = X509_STORE_CTX_get_error_depth( store );
		int err = X509_STORE_CTX_get_error( store );

		// Need to change logger
		sprintf( strBuf, "-Error with certificate at depth: %i\n", depth );
		d_cerr( strBuf );
		X509_NAME_oneline( X509_get_issuer_name( cert ), data, sizeof( data ) );
		sprintf( strBuf, " issuer = %s\n", data );
		d_cerr(strBuf);
		X509_NAME_oneline(X509_get_subject_name(cert), data, sizeof(data));
		sprintf(strBuf, " subject = %s\n", data);
		d_cerr(strBuf);
		sprintf(strBuf, " err %i:%s\n", err, X509_verify_cert_error_string(err));
		d_cerr(strBuf);
	}

	return ok;
}

DH* CSSL_CTX::CTX_DH_callback(
							 SSL* ssl, 
							 int is_export, 
							 int keylength)
{
	DH* ret;

	switch ( keylength )
	{
		case 512:
			ret = dh512;
			break;
		case 1024:
		default:
			ret = dh1024;
			break;
	}

	return ret;
}

bool CSSL_CTX::CTX_setup_DHparams(
								 const char* a_dh512File,
								 const char* a_dh1024File
								 )
{
	assert( a_dh512File );
	assert( a_dh1024File );

	BIO* bio;

	bio = BIO_new_file( a_dh512File, "r" );
	if ( !bio )
	{
		d_cerr("CTX_setup_DHparams: Error opening file dh512.pem");
		return false;
	}
	dh512 = PEM_read_bio_DHparams( bio, 0, 0, 0 );
	BIO_free(bio);
	if ( !dh512 )
	{
		d_cerr("CTX_setup_DHparams:  Error reading DH parameters from dh512.pem");
		return false;
	}

	bio = BIO_new_file(a_dh1024File, "r");
	if ( !bio )
	{
		d_cerr("CTX_setup_DHparams: Error opening file dh1024.pem");
		return false;
	}
	dh1024 = PEM_read_bio_DHparams( bio, 0, 0, 0 );
	BIO_free(bio);
	if ( !dh1024 )
	{
		d_cerr("CTX_setup_DHparams:  Error reading DH parameters from dh1024.pem");
		return false;
	}

	assert( ctx );
	if ( !ctx )
	{
		d_cerr("CTX_setup_DHparams:  SSL_CTX object is not created");
		return false;
	}
	SSL_CTX_set_tmp_dh_callback( ctx, CTX_DH_callback );

	return true;
}



bool CSSL_CTX::CTX_setup(
						SSL_METHOD* a_SSLmeth,
						const char* a_CertPath,
						const char* a_TrustedCert,
						const char* a_TrustedDir,
						const char* a_KeyPasswd,
						int a_verifyMode,
						int a_verifyDepth,
						long a_SSLoptions,
						const char* a_CipherList )
{
	assert( a_CertPath );
	assert( a_KeyPasswd );
	assert( a_CipherList );

	if ( ctx )
	{
		SSL_CTX_free(ctx);
		ctx = 0;
	};

	if ( !a_TrustedCert && !a_TrustedDir )
	{
		d_cerr("ssl_CTX_setup: error: CA file and CA directory both are NULL");
		return false;
	}

	//store passed arguments to have stable version
	m_CertPath = a_CertPath;

	if ( a_TrustedCert )
		m_TrustedCert = a_TrustedCert;
	else
		m_TrustedCert = "";

	if ( a_TrustedDir )
		m_TrustedDir = a_TrustedDir;
	else
		m_TrustedDir = "";

	m_KeyPasswd = a_KeyPasswd;
	m_CipherList = a_CipherList;

	//allocate SSL_CTX object
	ctx = SSL_CTX_new( a_SSLmeth );

	//if we have CA directory, use it
	//setup trusted certificates
	const char* CA_FILE = ( a_TrustedCert ) ? m_TrustedCert.c_str() : 0;
	const char* CA_DIR = ( a_TrustedDir ) ? m_TrustedDir.c_str() : 0;
	if ( SSL_CTX_load_verify_locations( ctx, CA_FILE, CA_DIR ) != 1 )
	{
		d_cerr("ssl_CTX_setup: Error loading CA file and/or directory");
		return false;
	}

	if ( SSL_CTX_set_default_verify_paths( ctx ) != 1 )
	{
		d_cerr( "Error loading default CA file and/or directory" );
		return false;
	}

	if ( SSL_CTX_use_certificate_chain_file( ctx, m_CertPath.c_str() ) != 1 )
	{
		d_cerr( "Error loading certificate from file" );
		return false;
	}

	//setup key encryption password
	SSL_CTX_set_default_passwd_cb( ctx, CTX_passwd_callback );
	SSL_CTX_set_default_passwd_cb_userdata( ctx, ( void* ) m_KeyPasswd.c_str() );

	if ( SSL_CTX_use_PrivateKey_file(ctx, m_CertPath.c_str(), SSL_FILETYPE_PEM) != 1 )
	{
		d_cerr( "Error loading private key from certificate file. Correct private key should be in the same cert file" );
		return false;
	}

	//set verify error dumping method
	SSL_CTX_set_verify( ctx, a_verifyMode, CTX_verify_callback );
	SSL_CTX_set_verify_depth( ctx, 4 );

	SSL_CTX_set_options( ctx, a_SSLoptions );

	if ( SSL_CTX_set_cipher_list( ctx, m_CipherList.c_str() ) != 1 )
	{
		d_cerr( "Error setting cipher list (no valid ciphers)" );
		return false;
	}

	return true;
}

	#ifdef __linux
void CSSL_CTX::sigPipeHandler( int )
{
		#ifdef _DEBUG
	cerr << "CSSL_CTX::sigPipeHandler: Broken pipe!" << endl;
		#endif // _DEBUG
}
	#endif // __linux

bool CSSL_CTX::init_OpenSSL()
{
	if ( !THREAD_setup() || ! SSL_library_init() )
	{
		d_cerr( "OpenSSL initialization failed!" );
		return false;
	}

	SSL_load_error_strings();

	SSL_seed_prng();

	#ifdef __linux
	// Catch possible SIGPIPE signal for application fail prevention
	signal( SIGPIPE, sigPipeHandler );
//    signal(SIGPIPE, SIG_IGN);
	#endif // __linux

	return true;
}

void CSSL_CTX::shutdown_OpenSSL()
{
	THREAD_cleanup();
}

int CSSL_CTX::THREAD_setup()
{
	int i;

	mutex_buf = ( MUTEX_TYPE* ) malloc( CRYPTO_num_locks() * sizeof( MUTEX_TYPE ) );

	if ( !mutex_buf )
		return 0;

	for ( i = 0; i < CRYPTO_num_locks(); i++ )
		MUTEX_SETUP( mutex_buf[ i ] );

	CRYPTO_set_id_callback( id_function );
	CRYPTO_set_locking_callback( locking_function );
	CRYPTO_set_dynlock_create_callback( dyn_create_function );
	CRYPTO_set_dynlock_lock_callback( dyn_lock_function );
	CRYPTO_set_dynlock_destroy_callback( dyn_destroy_function );

	return 1;
}

int CSSL_CTX::THREAD_cleanup()
{
	int i;

	if ( !mutex_buf )
		return 0;

	CRYPTO_set_id_callback( 0 );
	CRYPTO_set_locking_callback( 0 );
	CRYPTO_set_dynlock_create_callback( 0 );
	CRYPTO_set_dynlock_lock_callback( 0 );
	CRYPTO_set_dynlock_destroy_callback( 0 );

	for ( i = 0; i < CRYPTO_num_locks(); i++ )
		MUTEX_CLEANUP( mutex_buf[ i ] );

	free( mutex_buf );
	mutex_buf = 0;

	return 1;
}


SSL_CTX* CSSL_CTX::get()
{
	assert( ctx );
	return ctx;
}


///////////////////////////////////////////////////////////////////////
//
//                 CSSL class implementation
//
///////////////////////////////////////////////////////////////////////


CSSL::CSSL()    
{
	ssl = 0;
	bio = 0;
	lastErrno = 0;
	stillChanceToRead = true;
}

CSSL::~CSSL()   
{
	clean();
}

void CSSL::clean()
{
	if ( ssl )
	{
		SSL_shutdown( ssl );
		SSL_free( ssl );
		ssl = 0;
	}

	if ( bio )
	{
		BIO_free( bio );
		bio = 0;
	}
	lastErrno = 0;
	stillChanceToRead = true;
}


bool CSSL::setup_SSL_object(
						   SPTR_CSSL_CTX a_ptrCTX,
						   int a_fd,
						   SSL_objects_states a_obj_state
						   )
{
	assert( a_ptrCTX );
	int rc = 0;
	clean();

	if ( !a_ptrCTX )
	{
		d_cerr( "setup_SSL_object: CTX object not set" );
		return false;
	}

	ssl = SSL_new( a_ptrCTX->get() );
	if ( !ssl )
	{
		d_cerr( "setup_SSL_object: SSL_new error" );
		return false;
	}

	rc = SSL_set_fd( ssl, a_fd );
	if ( rc <= 0 )
	{
		d_cerr_sl( "setup_SSL_object: SSL_set_fd error " );
		lastErrno = rc = SSL_get_error( ssl, rc );
		d_cerr( rc );
		return false;
	}

	switch ( a_obj_state )
	{
		
		case SSL_OBJ_ACCEPT:
			SSL_set_accept_state( ssl );
			rc = ::SSL_accept( ssl );
			if ( rc <= 0 )
			{
				d_cerr_sl( "setup_SSL_object: SSL_accept error " );
				lastErrno = rc = SSL_get_error( ssl, rc );
				d_cerr( rc );
			}



			/*
			SSL_set_fd (ssl, fd);
					SSL_set_accept_state(ssl);
					if ((SSL_accept (ssl)) == -1)
	
	
	
			sprintf( strBuf, "%u", PORT++);
	
			bio = BIO_new_accept(strBuf);
			if (!bio)
			{
				d_cerr("SSL_OBJ_ACCEPT: Error creating server socket");
				return false;
			}
			d_cerr("SSL_OBJ_ACCEPT: creating server socket");
	
			if (BIO_do_accept(bio) <= 0)
			{
				d_cerr("SSL_OBJ_ACCEPT: Error binding server socket");
				return false;
			}
			d_cerr("SSL_OBJ_ACCEPT: binding server socket");
	
			if (BIO_do_accept(bio) <= 0)
			{
				d_cerr("SSL_OBJ_ACCEPT: Error accepting connection");
				return false;
			}
			d_cerr("SSL_OBJ_ACCEPT: accepting connection");
	
			client = BIO_pop(bio);
			d_cerr("SSL_OBJ_ACCEPT: BIO_pop");
	
			if (!(ssl = SSL_new(a_ctx_obj.ctx)))
			{
				d_cerr("SSL_OBJ_ACCEPT: Error creating SSL context");
				return false;
			}
			d_cerr("SSL_OBJ_ACCEPT: creating SSL context");
	
			SSL_set_accept_state(ssl);
			SSL_set_bio(ssl, client, client);
	
			if (::SSL_accept(ssl) <= 0)
			{
				d_cerr("SSL_OBJ_ACCEPT: Error accepting SSL connection");
				return false;
			}
			d_cerr("SSL_OBJ_ACCEPT: accepting SSL connection");
	*/

			break;

		case SSL_OBJ_CONNECT:
			SSL_set_connect_state( ssl );
			d_cerr( "SSL_connect started..." );
			rc = ::SSL_connect( ssl );
			if ( rc <= 0 )
			{
				d_cerr_sl( "setup_SSL_object: SSL_connect error " );
				lastErrno = rc = SSL_get_error( ssl, rc );
				d_cerr( rc );
			}

			/*        
			sprintf( strBuf, "localhost:%u", PORT++);
	
	
			bio = BIO_new_connect(strBuf);
			if (!bio)
			{
				d_cerr("SSL_OBJ_CONNECT: Error creating connection BIO");
				return false;
			}
			d_cerr("SSL_OBJ_CONNECT: creating connection BIO");
	
			if (BIO_do_connect(bio) <= 0)
			{
				d_cerr("SSL_OBJ_CONNECT: Error connecting to remote machine");
				return false;
			}
			d_cerr("SSL_OBJ_CONNECT: connecting to remote machine");
	
			if (!(ssl = SSL_new(a_ctx_obj.ctx)))
			{
				d_cerr("SSL_OBJ_CONNECT: Error creating SSL context");
				return false;
			}
			d_cerr("SSL_OBJ_CONNECT: creating SSL context");
			SSL_set_bio(ssl, bio, bio);
	
			if (::SSL_connect(ssl) <= 0)
			{
				d_cerr("SSL_OBJ_CONNECT: Error connecting SSL object");
				return false;
			}

			d_cerr("SSL_OBJ_CONNECT: connecting SSL object");
	*/

			break;

		default:
			assert( false );
			return false;
	}

	return true;
}


bool CSSL::SSL_accept(
					 SPTR_CSSL_CTX a_ptrCTX,
					 int a_fd
					 )
{
	d_cerr_sl( "SSL_accept: wrapping socket " );
	d_cerr( a_fd );

	if ( !setup_SSL_object( a_ptrCTX, a_fd, SSL_OBJ_ACCEPT ) )
	{
		return false;
	};

	return true;
}


bool CSSL::SSL_connect(
					  SPTR_CSSL_CTX a_ptrCTX,
					  int a_fd
					  )
{
	d_cerr_sl( "SSL_connect: wrapping socket " );
	d_cerr( a_fd );

	if ( !setup_SSL_object( a_ptrCTX, a_fd, SSL_OBJ_CONNECT ) )
	{
		return false;
	};

	return true;
}

int CSSL::SSL_read( void *buf, int num )
{
	if ( !ssl )
	{
		d_cerr( "SSL_read: ssl object not set" );
		return 0;
	}

	lastErrno = 0;
	int rc = 0;

	do
	{
		d_cerr_sl( "SSL_read: bytes " );
		d_cerr( num );

		rc = ::SSL_read( ssl, buf, num );

		if ( rc <= 0 )
		{
			lastErrno = SSL_get_error( ssl, rc );
			stillChanceToRead = false;
			d_cerr_sl( "SSL_read: error " );
			d_cerr( lastErrno );
		}
		else
		{
			//it might happen we can still read next time
			stillChanceToRead = true;
		}

	} while ( lastErrno == SSL_ERROR_WANT_READ ||
			  lastErrno == SSL_ERROR_WANT_WRITE
			);

	return rc;
}


int CSSL::SSL_write( const void *buf, int num )
{
	if ( !ssl )
	{
		d_cerr("SSL_write: ssl object not set");
		return 0;
	}

	lastErrno = 0;
	int rc = 0;

	do
	{
		d_cerr_sl( "SSL_write: bytes " );
		d_cerr( num );
		rc = ::SSL_write( ssl, buf, num );
		if ( rc <= 0 )
		{
			lastErrno = SSL_get_error( ssl,rc );
			d_cerr_sl( "SSL_write: error " );
			d_cerr( lastErrno );
		}

	} while ( lastErrno == SSL_ERROR_WANT_READ ||
			  lastErrno == SSL_ERROR_WANT_WRITE
			);

	return rc;
}


bool CSSL::good()
{
	if ( !ssl )
	{
		return false;
	}
	else
	{
		return true;
	}
}

bool CSSL::isChanceToRead()
{
	if ( !good() )
		return false;

	return stillChanceToRead;
}

#endif // USE_SSL

