/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// File: SSLcommon.h
// Author: Dmitry Ignatyuk
// Date: 16.10.04                                                                 
// Modified by:
// Last update:                                                                                     
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

#ifndef _SSLCOMMON_H_098098234526321
#define _SSLCOMMON_H_098098234526321

#include "openssl/bio.h"
#include "openssl/err.h"
#include "openssl/rand.h"
#include "openssl/ssl.h"
#include "openssl/x509v3.h"

#include <string>

#include "boost/shared_ptr.hpp"

#include <cassert>

#ifndef _WIN32
	#include <unistd.h>
	#include <pthread.h>
	#define THREAD_CC
	#define THREAD_TYPE pthread_t
	#define THREAD_CREATE( tid, entry, arg ) pthread_create( &( tid ), 0, \
                                        ( entry ), ( arg ) )
#else
	#include <windows.h>
	#define THREAD_CC __cdecl
	#define THREAD_TYPE DWORD
	#define THREAD_CREATE( tid, entry, arg ) do { \
	_beginthread( ( entry ), 0, ( arg ) ); \
	tid = GetCurrentThreadId(); \
	} while ( 0 )
#endif

#ifdef _WIN32
	#define MUTEX_TYPE HANDLE
	#define MUTEX_SETUP( x ) ( x ) = CreateMutex( 0, FALSE, 0 )
	#define MUTEX_CLEANUP( x ) CloseHandle( x )
	#define MUTEX_LOCK( x ) WaitForSingleObject( ( x ), INFINITE )
	#define MUTEX_UNLOCK( x ) ReleaseMutex( x )
	#define THREAD_ID GetCurrentThreadId()
#elif defined( _POSIX_THREADS )
	#define MUTEX_TYPE pthread_mutex_t
	#define MUTEX_SETUP( x ) pthread_mutex_init( &( x ), 0 )
	#define MUTEX_CLEANUP( x ) pthread_mutex_destroy( &( x ) )
	#define MUTEX_LOCK( x ) pthread_mutex_lock( &( x ) )
	#define MUTEX_UNLOCK( x ) pthread_mutex_unlock( &( x ) )
	#define THREAD_ID pthread_self()
#else
	#error You must define mutex operations appropriate for your platform!
#endif

struct CRYPTO_dynlock_value
{
	MUTEX_TYPE mutex;
};

class CSSL;

/**
 * class responsible for maintaining SSL context data
 */
class IPSOCKET_API CSSL_CTX
{
private:

	/**
	* Section for preventing creation of copy of CSSL_CTX
	*/
	CSSL_CTX( const CSSL_CTX& a_to_copy )
	{
		assert( false );
	};

	CSSL_CTX& operator=( const CSSL_CTX& a_to_copy )
	{
		assert( false ); 
		return *this;
	};

protected:
	friend class CSSL;

	/**
	 * service methods
	 */
	static CRYPTO_dynlock_value* dyn_create_function( const char* file, int line );
	static void locking_function( int mode, int n, const char* file, int line );
	static unsigned long id_function();
	static void dyn_lock_function( int mode, struct CRYPTO_dynlock_value* l, const char* file, int line );
	static void dyn_destroy_function( struct CRYPTO_dynlock_value* l, const char* file, int line );

	/**
	 * supply password for key encryption from userdata parameter
	 */
	static int CTX_passwd_callback(
								  char* buf, 
								  int size, 
								  int flag, 
								  void* userdata );

	/**
	 * prints error message in case if connection verification not passed
	 */
	static int CTX_verify_callback(
								  int ok, 
								  X509_STORE_CTX* store );

	/**
	 * DPparam methods and members block
	 */
	static DH* dh512;
	static DH* dh1024;
	static DH* CTX_DH_callback(
							  SSL* ssl, 
							  int is_export, 
							  int keylength ); 


	static MUTEX_TYPE* mutex_buf;
	static int SSL_seed_prng( int bytes = 10 );
	static int THREAD_setup();
	static int THREAD_cleanup();

#ifdef __linux
	static void sigPipeHandler( int );
#endif // __linux


	/**
	 * pointer to SSL context object
	 * !!!DON't use smartpointers, because special allocation functins involved
	 */
	SSL_CTX* ctx;

	std::string m_CertPath;
	std::string m_TrustedCert;
	std::string m_TrustedDir;
	std::string m_KeyPasswd;
	std::string m_CipherList;

public:
	CSSL_CTX();
	~CSSL_CTX();

	/**
	 * setup SSL context object
	 * 
	 * @param a_SSLmeth protocol version method   like result of
	 *              SSLv23_method()
	 * @param a_CertPath certficate of the side.
	 *           Should contain private key also in the file
	 *           file should be in PEM format
	 * @param a_TrustedCert path to file with certificate we can trust
	 *                      usually it's root cert or CA cert
	 *                      can be NULL if a_TrustedDir is not
	 * @param a_TrustedDir directory with trusted certificates
	 *                      can be NULL
	 * @param a_KeyPasswd password to use during keys encryption
	 * @param a_verifyMode certificate verification mode
	 * @param a_verifyDepth depth of chain for certification check
	 * @param a_SSLoptions options for autentification like
	 *                      SSL_OP_NO_SSLv2
	 * @param a_CipherList algorithms to use during autentification like
	 *                      "ALL:!ADH:!LOW:!EXP:!MD5:@STRENGTH"
	 * @return false -- setup was not successfull
	 *          true -- everything ok
	 * 
	 */ 
	bool CTX_setup(
				  SSL_METHOD* a_SSLmeth,
				  const char* a_CertPath,
				  const char* a_TrustedCert,
				  const char* a_TrustedDir,
				  const char* a_KeyPasswd,
				  int a_verifyMode,
				  int a_verifyDepth,
				  long a_SSLoptions,
				  const char* a_CipherList );


	/**
	 * setup Diffie-Helman parameter for encrypting keys
	 * 
	 * @param a_dh512File file containing 512 bit key
	 * @param a_dh1024File file containing 1024 bit key
	 * @return success state
	 */
	bool CTX_setup_DHparams(
						   const char* a_dh512File,
						   const char* a_dh1024File
						   );


	SSL_CTX* get();

	/**
	 * initializes all SSL libraries
	 */
	static bool init_OpenSSL();
	/**
	 * shutdown SSL sessions and everything
	 */
	static void shutdown_OpenSSL();

};

// Smart-pointer to CSSL_CTX object
typedef boost::shared_ptr < CSSL_CTX > SPTR_CSSL_CTX;


/**
 * wrapper over SSL object
 */
class CSSL
{
private:

	CSSL( const CSSL& a_to_copy )
	{
		assert( false );
	};

	CSSL& operator=( const CSSL& a_to_copy )
	{
		assert( false ); 
		return *this;
	};

protected:
	SSL* ssl;
	BIO* bio;
	int lastErrno;

	/**
	 * This variable defines lst state of read operation
	 * it might be that SSL has something in buffer select doesn't see
	 */
	bool stillChanceToRead;

	enum SSL_objects_states
	{
		SSL_OBJ_ACCEPT,
		SSL_OBJ_CONNECT
	};

	/**
	 * creates SSL object from existing SSL_CTX and TCP socked descriptor
	 * @param a_ctx_obj SSL CTX object qith context info
	 * @param a_fd TCP socket descriptor
	 * @param a_obj_state state of the SSL object (accept or connect)
	 * @return success status
	 */
	bool setup_SSL_object(
						 SPTR_CSSL_CTX a_ptrCTX,
						 int a_fd,
						 SSL_objects_states a_obj_state
						 );


public:

	CSSL();
	~CSSL();         

	/**
	 * creates SSL object from existing SSL_CTX and TCP socked descriptor
	 * @param a_ctx_obj SSL CTX object qith context info
	 * @param a_fd TCP socket descriptor
	 * @return success status
	 */
	bool SSL_connect(
					SPTR_CSSL_CTX a_ptrCTX,
					int a_fd
					);

	/**
	 * creates SSL object from existing SSL_CTX and TCP socked descriptor
	 * @param a_ctx_obj SSL CTX object qith context info
	 * @param a_fd TCP socket descriptor
	 * @return success status
	 */
	bool SSL_accept(
				   SPTR_CSSL_CTX a_ptrCTX,
				   int a_fd
				   );

	/**
	 * read information from SSL channel
	 * @param buf buffer to read to
	 * @param num amount of bytes to read
	 * @return >0 amount of byter to read, otherwise error
	 *         internal lastErrno will be raised in this case
	 */
	int SSL_read( void* buf, int num );

	/**
	 * write information to SSL channel
	 * @param buf buffer to read to
	 * @param num amount of bytes to read
	 * @return >0 amount of byter to read, otherwise error
	 *         internal lastErrno will be raised in this case
	 */
	int SSL_write( const void* buf, int num );


	void clean();

	/**
	 * returns status of SSL object
	 * @return is object ready to work
	 */
	bool good();


	/**
	 * @return if last read was successful and there is a chance to read more from SSL buffers
	 */
	bool isChanceToRead();

};


#endif //_SSLCOMMON_H_098098234526321

