/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// Created by: Sergey Belovolov                                                         
// Created on: 09.06.04                                                                 
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

#ifndef _IIPSTREAM_H
#define _IIPSTREAM_H

#include "IPSocket.h"

#include "ipstreambuf.h"

#include "boost_adopt/boost_adopt.h"
#include "hton/hton.h"

namespace IPStream
{
	// 100 milliseconds
	const unsigned int cuiRecvSleep = 100000000;

	// Input ipstream
	class iipstream : public virtual CIPSocket
	{
	private:
		// Buffer for receiving data
		ipstreambuf m_ipstrmBuffer;

		volatile bool m_bStopReceive;

		// Function for reducing number of codelines :)
		template < typename ReceivingType >
		inline void GuaranteedReceiving( ReceivingType& )
		{
			if ( m_bStopReceive || !good() )
				return;

			unsigned int uiSize = sizeof( ReceivingType );

			while ( !( ( m_ipstrmBuffer.Size() >= uiSize ) || m_bStopReceive ) && good() )
			{
				//operator>>( m_ipstrmBuffer );
				RecvBuffer( uiSize );
			} 
		}

		/**
		 * Method for receiving demanded amount of data
		 * 
		 * @param BuffSize size of demanded data in bytes.
		 * @return true if not less then BuffSize bytes are available in internal ipstreambuf.
		 */
		inline bool RecvBuffer( unsigned int BuffSize ) throw ( invalid_socket )
		{
			d_cerr_sl( "iipstream::RecvBuffer called with BuffSize: " );
			d_cerr( BuffSize );

			if ( m_bStopReceive || !BuffSize )
				return false;

			unsigned int uiInSize = m_ipstrmBuffer.Size();

#ifdef _SPLIT_OUT_BUF
			const unsigned int cuiRcvBufSize = _SPLIT_OUT_BUF;
#else
			const unsigned int cuiRcvBufSize = cuiBufSize;
#endif // _SPLIT_OUT_BUF

			// Temporary buffer. Used by "operator >> ( ipstreambuf& )"
			// Was moved here for more effitient working
			char cInBuf[ cuiRcvBufSize ];

#ifdef _IGNORE_BUFSIZE
			do
#else
			while ( !( ( uiInSize > ( BuffSize - 1 ) ) || m_bStopReceive ) && good() )
#endif // _IGNORE_BUFSIZE
			{
				int iRes = Recv( cInBuf, cuiRcvBufSize );
				d_cerr_sl( "iipstream::RecvBuffer Recv result: " );
				d_cerr( iRes );

				// If no data received
				if ( ciHasNoData == iRes )
				{
					// Sleep current thread for cuiRecvSleep milliseconds
					boost_adopt::sleep( 0, cuiRecvSleep );
					d_cerr( "iipstream::RecvBuffer no data sleep" );

					// Next loop
					continue;
				}

				// Connection was gracefully closed by peer or error happened
				if ( !iRes || ( iRes == -1 ) )
				{
					d_cerr( "iipstream::RecvBuffer : connection closed by peer or an error happened." );
					m_bStopReceive = true;
					break;
				}

				// Loop until we receive all data
				while ( ( iRes > 0 ) && !m_bStopReceive && good() )
				{
					// Put received
					m_ipstrmBuffer.Put( cInBuf, iRes );

					// Check if we've received last data portion
					if ( ( unsigned int ) iRes < cuiRcvBufSize )
						break;

					// Continue
					iRes = Recv( cInBuf, cuiRcvBufSize );
					d_cerr( "iipstream::RecvBuffer receiving all data" );

					// Connection was gracefully closed by peer 
					if ( !iRes )
					{
						m_bStopReceive = true;
						d_cerr( "iipstream::RecvBuffer : connection closed by peer." );
					}
				} // while ( ( iRes > 0 ) && !m_bStopReceive && good() )

				uiInSize = m_ipstrmBuffer.Size();

			} // do
#ifdef _IGNORE_BUFSIZE
			while ( !( ( uiInSize > ( BuffSize - 1 ) ) || m_bStopReceive ) && good() );
#endif // _IGNORE_BUFSIZE

			d_cerr( "" );

			if ( uiInSize > ( BuffSize - 1 ) )
				return true;
			else if ( !good() )
				throw invalid_socket( "iipstream::RecvBuffer: IPSocket become invalid." );

			return false;
		}

	public:

		// Drop all the stored at the buffer information
		inline void Clear()
		{
			d_cerr( "iipstream::Clear called." );
			d_cerr( "" );

			m_ipstrmBuffer.Clear();
		}

		// Functor for catching signal of stream termination
		inline void operator()()
		{
			d_cerr( "iipstream::operator() called." );
			d_cerr( "" );

			m_bStopReceive = true;
		}

		// Default constructor
		inline iipstream( unsigned short PortNumber = 0, 
						  const std::string& address = "",
						  ETransportProtocol protocol = TP_TCP, 
						  int options = 0, 
						  EDomain domain = D_INET 
						) : 
		CIPSocket( PortNumber, address, protocol, options, domain ), m_bStopReceive( false )
		{
			d_cerr( "iipstream constructor called." );
			d_cerr( "" );
		}

#ifdef USE_SSL
		// SSL constructor
		inline iipstream( SPTR_CSSL_CTX a_ptrCTX,
						  unsigned short PortNumber, 
						  const std::string& address,
						  int options = 0, 
						  EDomain domain = D_INET 
						) : 
		CIPSocket( a_ptrCTX,
				   PortNumber, 
				   address, 
				   options, 
				   domain ), m_bStopReceive( false )
		{
			d_cerr( "SSL iipstream constructor called." );
			d_cerr( "" );
		}
#endif

		// For construction via Accept
		inline iipstream( const CIPSocket& IPS ) : 
		CIPSocket( IPS ), m_bStopReceive( false )
		{
			d_cerr( "iipstream constructor via Accept called." );
			d_cerr( "" );
		}

		inline ~iipstream()
		{
			d_cerr( "iipstream destructor called." );
			d_cerr( "" );

			m_bStopReceive = true;
		}

		// For avoiding call of template operator>>
		inline iipstream& operator >> ( std::string& Object ) 
			throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( std::string& ) called." );
			d_cerr( "" );

			Object = "";

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "iipstream::operator >> ( std::string& ): IPSocket become invalid." );
				//return *this;

			// Variable for storing sizes of types
			unsigned int uiSize = sizeof( unsigned int );

			while ( m_ipstrmBuffer.Size() < uiSize && !m_bStopReceive && good() )
			{
				//operator>>( m_ipstrmBuffer );
				RecvBuffer( uiSize );
			}

			// Length of following string
			unsigned int uiStringLength = 0;

			if ( m_bStopReceive || !good() )
			{
				d_cerr( "iipstream::operator >> ( std::string& ) channel dropped during string receiving." );
				return *this;
			}

			// Read string length
			assert( m_ipstrmBuffer.Size() >= uiSize );

			memcpy( &uiStringLength, m_ipstrmBuffer.Get(), uiSize );

			// Total length of string buffer 
			unsigned int uiBufLength = uiSize + ntohl( uiStringLength );

			while ( ( m_ipstrmBuffer.Size() < uiBufLength ) && !m_bStopReceive && good() )
			{
				//operator>>( m_ipstrmBuffer );
				RecvBuffer( uiBufLength );
			}

			if ( m_bStopReceive || !good() )
			{
				d_cerr( "iipstream::operator >> ( std::string& ) channel dropped during string receiving" );
				return *this;
			}

			// Check if full string was received
			if ( !( m_ipstrmBuffer.Size() < uiBufLength ) )
				m_ipstrmBuffer >> Object;

			return *this;
		}

		// Realization for long (and int in I32 architecture)
		inline iipstream& operator >> ( unsigned long long& ForInput )
			throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( unsigned long long& ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "iipstream::operator >> ( unsigned long long& ): IPSocket become invalid." );
				//return *this;

			GuaranteedReceiving( ForInput );

			if ( !( m_ipstrmBuffer.Empty() || ( m_ipstrmBuffer.Size() < sizeof( ForInput ) ) ) )
			{
				m_ipstrmBuffer >> ForInput;

				ForInput = ntohll( ForInput );
			}

			return *this;
		}

		inline iipstream& operator >> ( long long& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( long long& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned long long& ) ForInput );
		}

		// Realization for long (and int in I32 architecture)
		inline iipstream& operator >> ( unsigned long& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( unsigned long& ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "iipstream::operator >> ( unsigned long& ): IPSocket become invalid." );
				//return *this;

			GuaranteedReceiving( ForInput );

			if ( !( m_ipstrmBuffer.Empty() || ( m_ipstrmBuffer.Size() < sizeof( ForInput ) ) ) )
			{
				m_ipstrmBuffer >> ForInput;

				ForInput = ntohl( ForInput );
			}

			return *this;
		}

		inline iipstream& operator >> ( long& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( long& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned long& ) ForInput );
		}

		inline iipstream& operator >> ( unsigned int& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( unsigned int& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned long& ) ForInput );
		}

		inline iipstream& operator >> ( int& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( int& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned long& ) ForInput );
		}

		// Realization for short
		inline iipstream& operator >> ( unsigned short& ForInput )
		{
			d_cerr( "iipstream::operator >> ( unsigned short& ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "iipstream::operator >> ( unsigned short& ): IPSocket become invalid." );
				//return *this;

			GuaranteedReceiving( ForInput );

			if ( !( m_ipstrmBuffer.Empty() || ( m_ipstrmBuffer.Size() < sizeof( ForInput ) ) ) )
			{
				m_ipstrmBuffer >> ForInput;

				ForInput = ntohs( ForInput );
			}

			return *this;
		}

		inline iipstream& operator >> ( short& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( short& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned short& ) ForInput );
		}

		// Realization for char
		inline iipstream& operator >> ( unsigned char& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( unsigned char& ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "iipstream::operator >> ( unsigned char& ): IPSocket become invalid." );
				//return *this;

			GuaranteedReceiving( ForInput );

			if ( !( m_ipstrmBuffer.Empty() || ( m_ipstrmBuffer.Size() < sizeof( ForInput ) ) ) )
				m_ipstrmBuffer >> ForInput;

			return *this;
		}

		inline iipstream& operator >> ( char& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( char& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned char& ) ForInput );
		}

		inline iipstream& operator >> ( bool& ForInput ) throw ( invalid_socket )
		{
			d_cerr( "iipstream::operator >> ( bool& ) called." );
			d_cerr( "" );

			return operator>>( ( unsigned char& ) ForInput );
		}

		// Realization for ipstreambuf
		inline iipstream& operator >> ( ipstreambuf& ForInput )
		{
			d_cerr( "iipstream::operator >> ( ipstreambuf& ) called." );

			if ( m_bStopReceive )
				return *this;

			unsigned int uiInSize = ForInput.Size();

#ifdef _SPLIT_OUT_BUF
			const unsigned int cuiRcvBufSize = _SPLIT_OUT_BUF;
#else
			const unsigned int cuiRcvBufSize = cuiBufSize;
#endif // _SPLIT_OUT_BUF

			// Temporary buffer. Used by "operator >> ( ipstreambuf& )"
			// Was moved here for more effitient working
			char cInBuf[ cuiRcvBufSize ];

#ifdef _IGNORE_BUFSIZE
			do
#else
			while ( !( ( uiInSize > 0 ) || m_bStopReceive ) && good() )
#endif // _IGNORE_BUFSIZE
			{
				int iRes = Recv( cInBuf, cuiRcvBufSize );
				d_cerr_sl( "iipstream::operator >> ( ipstreambuf& ) Recv result: " );
				d_cerr( iRes );

				// If no data received
				if ( ciHasNoData == iRes )
				{
					// Sleep current thread for cuiRecvSleep milliseconds
					boost_adopt::sleep( 0, cuiRecvSleep );
					d_cerr( "iipstream::operator >> ( ipstreambuf& ) no data sleep" );

					// Next loop
					continue;
				}

				// Connection was gracefully closed by peer or error happened
				if ( !iRes || ( iRes == -1 ) )
				{
					d_cerr( "iipstream::operator >> ( ipstreambuf& ) : connection closed by peer or an error happened." );
					m_bStopReceive = true;
					break;
				}

				// Loop until we receive all data
				while ( ( iRes > 0 ) && !m_bStopReceive && good() )
				{
					// Put received
					ForInput.Put( cInBuf, iRes );

					// Check if we've received last data portion
					if ( (unsigned int)iRes < cuiRcvBufSize )
						break;

					// Continue
					iRes = Recv( cInBuf, cuiRcvBufSize );
					d_cerr( "iipstream::operator >> ( ipstreambuf& ) receiving all data" );

					// Connection was gracefully closed by peer 
					if ( !iRes )
					{
						m_bStopReceive = true;
						d_cerr( "iipstream::operator >> ( ipstreambuf& ) : connection closed by peer." );
					}
				}

				uiInSize = ForInput.Size();

			} // do
#ifdef _IGNORE_BUFSIZE
			while ( !( ( uiInSize > 0 ) || m_bStopReceive ) && good() );
#endif // _IGNORE_BUFSIZE

			d_cerr( "" );

			return *this;
		}

	};

} // namespace IPStream

#endif // _IIPSTREAM_H

