/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// Created by: Sergey Belovolov                                                         
// Created on: 10.06.04                                                                 
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

#ifndef _IPSTREAMBUF_H
#define _IPSTREAMBUF_H

#include <vector>
#include <stdexcept>

#include "hton/hton.h"

namespace IPStream
{
	// Default size of the buffer (0x400 - 1Kb, 0x600 - 1.5Kb)
	const unsigned int cuiBufSize = 0x1200;//0x600;

	// Exception for invalid reading from ipstreambuf
	class except_e_size : public std::runtime_error
	{
	public:
		except_e_size( const std::string& message = "Not enough data in the buffer." ) :
		std::runtime_error( message ) 
		{
		}
	};

	// Auxilary class for simplifying work with ipstreams. Not thread-safe!
	class ipstreambuf
	{
	private:
		// Define type of the buffer
		typedef std::vector < char > BuffType;

		// Internal buffer
		BuffType m_vcBuf;

		// Action on logic error
		inline void OnSizeError( const std::string& message = "Not enough data in the buffer." ) const 
			throw( except_e_size )
		{
			throw except_e_size( message );
		}

	public:

		
		// Default constructor
		inline ipstreambuf()
		{
			// Prepare space for buffer
			//m_vcBuf.reserve( cuiBufSize );
		}
		

		// Interval constructor
		template< typename InputIterator >
		inline ipstreambuf( InputIterator First, InputIterator Last ) :
		m_vcBuf( First, Last )
		{
		}

		// Put binary data into the buffer
		inline bool Put( const char* Buf = 0, unsigned int size = 0 )
		{
			// Check passed values
			if ( !( Buf && size ) )
				return false;

			// Copy object into the end of the buffer
			//std::copy( Buf, Buf + size, std::back_inserter( m_vcBuf ) );
			m_vcBuf.insert( m_vcBuf.end(), Buf, Buf + size );

			return true;
		}

		// Get binary data from the buffer (and remove it from buffer)
		inline bool Get( char* Buf, unsigned int size = 0 )
		{
			// Check passed values
			if ( !( Buf && size ) )
				return false;

			// If we have enough quantity in buffer
			if ( Size() >= size )
			{
				// Copy data from the beginning of the buffer into passed object
				memcpy( Buf, &m_vcBuf[ 0 ], size );

				// Drop read data from the beginning of the buffer
				BuffType( m_vcBuf.begin() + size, m_vcBuf.end() ).swap( m_vcBuf );
			}
			else
				OnSizeError( "Get: Not enough data in the buffer." );

			return true;
		}

		// Get pointer to the buffer beginning
		inline const char* Get() const
		{
			if ( !Size() )
				return 0;

			return &m_vcBuf[ 0 ];
		}

///////////////////////

		// Realization for long long
		inline ipstreambuf& operator << ( unsigned long long ForOutput )
		{
			//ForOutput = htonll( ForOutput );

			// N.B.: It would be not bad to check the result
			Put( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline ipstreambuf& operator << ( long long ForOutput )
		{
			return operator<<( ( unsigned long long ) ForOutput );
		}

		// Realization for long (and int in I32 architecture)
		inline ipstreambuf& operator << ( unsigned long ForOutput )
		{
			//ForOutput = htonl( ForOutput );

			Put( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline ipstreambuf& operator << ( long ForOutput )
		{
			return operator<<( ( unsigned long ) ForOutput );
		}

		inline ipstreambuf& operator << ( unsigned int ForOutput )
		{
			return operator<<( ( unsigned long ) ForOutput );
		}

		inline ipstreambuf& operator << ( int ForOutput )
		{
			return operator<<( ( unsigned long ) ForOutput );
		}

		// Realization for short
		inline ipstreambuf& operator << ( unsigned short ForOutput )
		{
			//ForOutput = htons( ForOutput );

			Put( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline ipstreambuf& operator << ( short ForOutput )
		{
			return operator<<( ( unsigned short ) ForOutput );
		}

		// Realization for char
		inline ipstreambuf& operator << ( unsigned char ForOutput )
		{
			Put( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline ipstreambuf& operator << ( char ForOutput )
		{
			return operator<<( ( unsigned char ) ForOutput );
		}

		inline ipstreambuf& operator << ( bool ForOutput )
		{
			return operator<<( ( unsigned char ) ForOutput );
		}

		// Realization for long (and int in I32 architecture)
		inline ipstreambuf& operator >> ( unsigned long long& ForInput )
		{
			Get( ( char* ) &ForInput, sizeof( ForInput ) );

            //ForInput = ntohll( ForInput );
			
			return *this;
		}

		inline ipstreambuf& operator >> ( long long& ForInput )
		{
			return operator>>( ( unsigned long long& ) ForInput );
		}

		// Realization for long (and int in I32 architecture)
		inline ipstreambuf& operator >> ( unsigned long& ForInput )
		{
			Get( ( char* ) &ForInput, sizeof( ForInput ) );

			//ForInput = ntohl( ForInput );
			
			return *this;
		}

		inline ipstreambuf& operator >> ( long& ForInput )
		{
			return operator>>( ( unsigned long& ) ForInput );
		}

		inline ipstreambuf& operator >> ( unsigned int& ForInput )
		{
			return operator>>( ( unsigned long& ) ForInput );
		}

		inline ipstreambuf& operator >> ( int& ForInput )
		{
			return operator>>( ( unsigned long& ) ForInput );
		}

		// Realization for short
		inline ipstreambuf& operator >> ( unsigned short& ForInput )
		{
			Get( ( char* ) &ForInput, sizeof( ForInput ) );

			//ForInput = ntohs( ForInput );

			return *this;
		}

		inline ipstreambuf& operator >> ( short& ForInput )
		{
			return operator>>( ( unsigned short& ) ForInput );
		}

		// Realization for char
		inline ipstreambuf& operator >> ( unsigned char& ForInput )
		{
			Get( ( char* ) &ForInput, sizeof( ForInput ) );

			return *this;
		}

		inline ipstreambuf& operator >> ( char& ForInput )
		{
			return operator>>( ( unsigned char& ) ForInput );
		}

		inline ipstreambuf& operator >> ( bool& ForInput )
		{
			return operator>>( ( unsigned char& ) ForInput );
		}

///////////////////////

		// Get current buffer size
		inline unsigned int Size() const
		{
			return m_vcBuf.size();
		}

		// Get current buffer size
		inline unsigned int Empty() const
		{
			return m_vcBuf.empty();
		}

		// Clear the buffer (erase all its contents)
		inline void Clear()
		{
			return m_vcBuf.clear();
		}

		inline ipstreambuf& operator << ( const char* Object )
		{
			unsigned int uiStrLength = strlen( Object ) + 1;

			// Send the string length
			operator<<( ( unsigned long ) htonl( uiStrLength ) );

			// Copy object into the end of the buffer
			Put( Object, uiStrLength );

			return *this;
		}

		inline ipstreambuf& operator << ( const std::string& Object )
		{
			return operator<<( Object.c_str() );
		}
		
		// Works with NULL-terminated strings in buffer only!!!
		inline ipstreambuf& operator >> ( std::string& Object )
		{
            unsigned int uiBufSize = Size();

			// Check if we have data
			if ( uiBufSize > sizeof( unsigned int ) )
			{
				unsigned int uiStrLength = 0;

				Get( ( char* ) &uiStrLength, sizeof( unsigned int ) );

				uiStrLength = ntohl( uiStrLength );

				// If string is wrong
				if ( !uiStrLength )
					return *this;

				const char* cpTmp = Get();

				unsigned int uiStrSize = strlen( cpTmp );               

                // This won't work with debug STLPort 
				std::string strTemp( cpTmp, cpTmp + 
									 ( ( uiStrSize < uiBufSize ) ? uiStrSize : uiBufSize ) );

				Object = strTemp;

				// Delete read data from the buffer
				// Watch out - we write to container buffer directly!
				Get( &strTemp[ 0 ], ( ( ( strTemp.length() + 1 ) < uiBufSize ) ? 
									  ( strTemp.length() + 1 ) : uiBufSize ) );

				// Check if string was received correctly
				if ( Object.length() != ( uiStrLength - 1 ) )
					Object = "";

			}
			else
				// Always check the size of buffer content before reading or 
				// catch this exception
				OnSizeError( "operator>>( std::string& ): Not enough data in the buffer." );

            return *this;
		}
	};
} // namespace IPStream

#endif // _IPSTREAMBUF_H
