/////////////////////////////////////////////////////////////////////////////           
//                                                                                      
// Created by: Sergey Belovolov                                                         
// Created on: 09.06.04                                                                 
//                                                                                      
/////////////////////////////////////////////////////////////////////////////           
// (C) 2004 Sergey Belovolov                                           
// e-mail: b_sv@mail.ru
/////////////////////////////////////////////////////////////////////////////

#ifndef _OIPSTREAM_H
#define _OIPSTREAM_H

#include "IPSocket.h"

#include "ipstreambuf.h"

#ifdef _HAS_SEND_BUFFER
	#include "boost/thread/condition.hpp"
	#include "boost/thread/mutex.hpp"
	#include "boost/thread/thread.hpp"
	#include "boost/shared_ptr.hpp"
#endif // _HAS_SEND_BUFFER

#include "boost_adopt/boost_adopt.h"
#include "hton/hton.h"

namespace IPStream
{
// If _HAS_SEND_BUFFER defined then send operation will be prebuffered
#ifdef _HAS_SEND_BUFFER
	// 10 milliseconds
	const unsigned int cuiSendSleep = 10000000;
#endif // _HAS_SEND_BUFFER

	// Output ipstream
	class oipstream : public virtual CIPSocket
	{
#ifdef _HAS_SEND_BUFFER
	private:
		// Buffer for sending data
		ipstreambuf m_ipstrmBuffer;

		// Flag for sending thread
		volatile bool m_bStopSend;

		// Mutex for controlling access to input buffer
		BRMutex m_mtxBufAccess;

		// Condition for controlling minimum size of input buffer
		boost::condition m_cndBufEmpty; 

		// Thread for sending data via buffer m_ipstrmBuffer
		SPThread m_spthrdSendThread;
#endif // _HAS_SEND_BUFFER
	public:

		// Default constructor
		inline oipstream( unsigned short PortNumber = 0, 
						  const std::string& address = "",
						  ETransportProtocol protocol = TP_TCP, 
						  int options = 0, 
						  EDomain domain = D_INET 
                          ) : 
		CIPSocket( PortNumber, address, protocol, options, domain 
                   )
#ifdef _HAS_SEND_BUFFER
		, m_bStopSend( false ),
			m_spthrdSendThread( new boost::thread(
												 boost_adopt::class_member_2_thread_adapter 
												 < oipstream, void* > 
												 ( this, &oipstream::SendThread, 0 ) ) )
#endif // _HAS_SEND_BUFFER
			
		{
			d_cerr( "oipstream constructor called." );
			d_cerr( "" );
		}

#ifdef USE_SSL
        // SSL constructor
        inline oipstream( SPTR_CSSL_CTX a_ptrCTX,
                          unsigned short PortNumber, 
                          const std::string& address,
                          int options = 0, 
                          EDomain domain = D_INET 
                          ) : 
        CIPSocket( a_ptrCTX,
                   PortNumber, 
                   address, 
                   options, 
                   domain 
                   )
#ifdef _HAS_SEND_BUFFER
        , m_bStopSend( false ),
            m_spthrdSendThread( new boost::thread(
                                                 boost_adopt::class_member_2_thread_adapter 
                                                 < oipstream, void* > 
                                                 ( this, &oipstream::SendThread, 0 ) ) )
#endif // _HAS_SEND_BUFFER

        {
            d_cerr( "SSL oipstream constructor called." );
            d_cerr( "" );
        }
#endif //USE_SSL


		// For construction via Accept
		inline oipstream( const CIPSocket& IPS ) : 
		CIPSocket( IPS )
#ifdef _HAS_SEND_BUFFER
		, m_bStopSend( false ),
			m_spthrdSendThread( new boost::thread(
												 boost_adopt::class_member_2_thread_adapter 
												 < oipstream, void* > 
												 ( this, &oipstream::SendThread, 0 ) ) )
#endif // _HAS_SEND_BUFFER
			
		{
			d_cerr( "oipstream constructor via Accept called." );
			d_cerr( "" );
		}

#ifdef _HAS_SEND_BUFFER
		// For copy construction
		inline oipstream( const oipstream& IPStrm )	/*: CIPSocket( IPStrm )*/
		{
			d_cerr( "oipstream copy constructor called." );
			d_cerr( "" );

			// Call operator=
			*this = IPStrm;
		}

		inline oipstream& operator=( const oipstream& IPStrm )
		{
			d_cerr( "oipstream::operator= ( const oipstream& ) called." );
			d_cerr( "" );

			if ( this != &IPStrm )
			{
				m_ipstrmBuffer = IPStrm.m_ipstrmBuffer;

				// For thread stopping
				m_bStopSend = true;

				// If thread is waiting
				m_cndBufEmpty.notify_one(); 

				// If thread was created
				if ( m_spthrdSendThread.get() )
					// Wait untill the thread stops
					m_spthrdSendThread -> join();

				m_bStopSend = false;

				m_spthrdSendThread.reset( new boost::thread(
														   boost_adopt::class_member_2_thread_adapter 
														   < oipstream, void* > 
														   ( this, &oipstream::SendThread, 0 ) ) );
			}

			return *this;
		}

		inline ~oipstream()
		{
			d_cerr( "oipstream destructor called." );
			d_cerr( "" );

			// For thread stopping
			m_bStopSend = true;

			// If thread is waiting
			m_cndBufEmpty.notify_one(); 

			// If thread was created
			if ( m_spthrdSendThread.get() )
				// Wait untill the thread stops
				m_spthrdSendThread -> join();
		}

		// Thread of deffered data sending
		inline void SendThread( void* )
		{
			d_cerr( "oipstream::SendThread started." );
			d_cerr( "" );

			// Main thread starts
			while ( !m_bStopSend && good() )
			{
				// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
				BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess ); 

				// Check if miniimum size of queue is achieved then wait before next step
				while ( !( m_ipstrmBuffer.Size() || m_bStopSend ) )
					m_cndBufEmpty.wait( scoped_lock_buffer ); 

				// Time to exit
				if ( m_bStopSend )
					break;

				// If buffer size is not suffitient for immediate sending
				if ( m_ipstrmBuffer.Size() < cuiBufSize )
				{
					scoped_lock_buffer.unlock();

					// Give processor time to other processes
					boost_adopt::sleep( 0, cuiSendSleep );

					scoped_lock_buffer.lock();
				}

				// Flash();

				// N.B.: It would be not bad to check the result
				Send( m_ipstrmBuffer.Get(), m_ipstrmBuffer.Size() );

				// Drop the data that was sent
				m_ipstrmBuffer.Clear();
			} // while ( !m_bStopSend && good() )
		}
#endif // _HAS_SEND_BUFFER

/*
		// Template for data outputting into the ipstream
		// Conversion between little- and big-endian of a non-basic SerialType
		// is a headache of yours 
		template < typename SerialType >
		inline oipstream& operator << ( const SerialType& Object )
		{
			d_cerr( "oipstream::operator << ( const SerialType& ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				return *this;

#ifdef _HAS_SEND_BUFFER
			// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
			BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

			m_ipstrmBuffer << Object;

			operator<<( m_ipstrmBuffer );
#else
			// Temporary intermediate buffer
			ipstreambuf ipstrmOutBuf;

			ipstrmOutBuf << Object;

			operator<<( ipstrmOutBuf );
#endif // _HAS_SEND_BUFFER

			return *this;
		}
*/

		// Realization for long long
		inline oipstream& operator << ( unsigned long long ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( unsigned long long ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "oipstream::operator << ( unsigned long long ): IPSocket become invalid." );
				//return *this;

			ForOutput = htonll( ForOutput );
#ifdef _HAS_SEND_BUFFER
			// If object size is too small
			if ( ( sizeof( ForOutput ) < cuiBufSize ) || !m_ipstrmBuffer.Empty() )
			{
				// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
				BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

				m_ipstrmBuffer << ForOutput;

				operator<<( m_ipstrmBuffer );
			}
			else
#endif // _HAS_SEND_BUFFER
				// N.B.: It would be not bad to check the result
				Send( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline oipstream& operator << ( long long ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( long long ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned long long ) ForOutput );
		}

		// Realization for long (and int in I32 architecture)
		inline oipstream& operator << ( unsigned long ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( unsigned long ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "oipstream::operator << ( unsigned long ): IPSocket become invalid." );
				//return *this;

			ForOutput = htonl( ForOutput );
#ifdef _HAS_SEND_BUFFER
			// If object size is too small
			if ( ( sizeof( ForOutput ) < cuiBufSize ) || !m_ipstrmBuffer.Empty() )
			{
				// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
				BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

				m_ipstrmBuffer << ForOutput;

				operator<<( m_ipstrmBuffer );
			}
			else
#endif // _HAS_SEND_BUFFER
				// N.B.: It would be not bad to check the result
				Send( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline oipstream& operator << ( long ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( long ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned long ) ForOutput );
		}

		inline oipstream& operator << ( unsigned int ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( unsigned int ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned long ) ForOutput );
		}

		inline oipstream& operator << ( int ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( int ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned long ) ForOutput );
		}

		// Realization for short
		inline oipstream& operator << ( unsigned short ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( unsigned short ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "oipstream::operator << ( unsigned short ): IPSocket become invalid." );
				//return *this;

			ForOutput = htons( ForOutput );
#ifdef _HAS_SEND_BUFFER
			// If object size is too small
			if ( ( sizeof( ForOutput ) < cuiBufSize ) || !m_ipstrmBuffer.Empty() )
			{
				// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
				BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

				m_ipstrmBuffer << ForOutput;

				operator<<( m_ipstrmBuffer );
			}
			else
#endif // _HAS_SEND_BUFFER
				// N.B.: It would be not bad to check the result
				Send( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline oipstream& operator << ( short ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( short ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned short ) ForOutput );
		}

		// Realization for char
		inline oipstream& operator << ( unsigned char ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( unsigned char ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "oipstream::operator << ( unsigned char ): IPSocket become invalid." );
				//return *this;

#ifdef _HAS_SEND_BUFFER
			// If object size is too small
			if ( ( sizeof( ForOutput ) < cuiBufSize ) || !m_ipstrmBuffer.Empty() )
			{
				// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
				BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

				m_ipstrmBuffer << ForOutput;

				operator<<( m_ipstrmBuffer );
			}
			else
#endif // _HAS_SEND_BUFFER
				// N.B.: It would be not bad to check the result
				Send( ( const char* ) &ForOutput, sizeof( ForOutput ) );

			return *this;
		}

		inline oipstream& operator << ( char ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( char ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned char ) ForOutput );
		}

		inline oipstream& operator << ( bool ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( bool ) called." );
			d_cerr( "" );

			return operator<<( ( unsigned char ) ForOutput );
		}

		// Realization for C-style strings
		inline oipstream& operator << ( const char* ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( const char* ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "oipstream::operator << ( const char* ): IPSocket become invalid." );
				//return *this;

#ifdef _HAS_SEND_BUFFER
			// If object size is too small
			if ( ( ( strlen( ForOutput ) + 1 ) < cuiBufSize ) || !m_ipstrmBuffer.Empty() )
			{
				// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
				BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

				m_ipstrmBuffer << ForOutput;

				operator<<( m_ipstrmBuffer );
			}
			else
#endif // _HAS_SEND_BUFFER
			{
				// Put the string length with null-terminator
				operator<<( ( unsigned int ) ( strlen( ForOutput ) + 1 ) );

				// N.B.: It would be not bad to check the result
				Send( ForOutput, strlen( ForOutput ) + 1 );
			}
                
			return *this;
		}

		// Realization for std strings
		inline oipstream& operator << ( const std::string& ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( const std::string& ) called." );
			d_cerr( "" );

			return operator<<( ForOutput.c_str() );
		}

		// Realization for std strings
		inline oipstream& operator << ( const ipstreambuf& ForOutput ) throw ( invalid_socket )
		{
			d_cerr( "oipstream::operator << ( const ipstreambuf& ) called." );
			d_cerr( "" );

			// Check if stream is invalid
			if ( !good() )
				throw invalid_socket( "oipstream::operator << ( const ipstreambuf& ): IPSocket become invalid." );
				//return *this;

#ifdef _HAS_SEND_BUFFER
			// If we work with internal buffer
			if ( &ForOutput == &m_ipstrmBuffer )
			{
				// If thread is waiting
				m_cndBufEmpty.notify_one();
			}
			else // External buffer
			{
				// Not suffitient size
				if ( ( ForOutput.Size() < cuiBufSize ) || !m_ipstrmBuffer.Empty() )
				{
					// Set lock for avoiding simultaneous access to m_ipstrmBuffer 
					BRMutex::scoped_lock scoped_lock_buffer( m_mtxBufAccess );

					// Copy into internal buffer
					m_ipstrmBuffer.Put( ForOutput.Get(), ForOutput.Size() );

					scoped_lock_buffer.unlock();

					m_cndBufEmpty.notify_one();
				}
				else
#endif // _HAS_SEND_BUFFER
					// N.B.: It would be not bad to check the result
					Send( ForOutput.Get(), ForOutput.Size() );
#ifdef _HAS_SEND_BUFFER
			}
#endif // _HAS_SEND_BUFFER
			return *this;
		}
	};
} // namespace IPStream

#endif // _OIPSTREAM_H

